#!/usr/bin/env python

__id__ = "$Id: trio_test.py 801 2008-10-28 22:22:46Z hyunjung $"

from getpass import getpass
import optparse
import signal
import sys
from trio import triodb
from trio import common
import ConfigParser
import os

TESTROOT = os.path.join(common.ROOT, "..", "test")
test_conf = ConfigParser.ConfigParser()
test_conf.read(os.path.join(TESTROOT, "triotest.cfg"))

def print_action(msg):
    print "%s" % (msg + "...").ljust(60),

def print_success():
    print "SUCCESS!"

def print_failure():
    print "FAILURE!"

def initdb():
    print_action("Initializing database %s" % test_conf.get("global", "dbname"))
    os.system("dropdb %s" % test_conf.get("global", "dbname"))
    if os.system("createdb %s" % test_conf.get("global", "dbname")) != 0:
        raise Exception("Couldn't create database %s" % test_conf.get("global", "dbname"))
    if os.system("%s %s" % ("triosetupdb", test_conf.get("global", "dbname"))) != 0:
        raise Exception("Couldn't create database %s" % test_conf.get("global", "dbname"))
    print_success()

def xtuples2str(xtuples):
    s = [str(x) for x in xtuples]
    s.sort()
    val = ""
    for xtuple in s:
        val += xtuple
    return val

def execute_file(cnx, file, outfile=None):
    c = cnx.cursor()
    commands = ""
    if outfile is None:
        file = os.path.join(TESTROOT, file)
    else:
        s = ""
    f = open(file)
    for line in f:
        if not line.strip().startswith("--"):
            commands += line
        parts = commands.split(";")
        commands = parts[-1:][0]
        comm_list = parts[:-1]
        for comm in comm_list:
            print comm
            try:
                c.execute(comm)
                if outfile is not None:
                    s += xtuples2str(c.xfetchall())
                    s += "\n"
                cnx.commit()
            except Exception, ex:
                common.print_exception(ex)
                print_failure()
                print("\n################################################")
                print comm
                print("################################################\n")
                raise Exception(comm)
    if outfile is not None:
        outfile.write(s)

def test_file(cnx, file):
    c = cnx.cursor()
    commands = ""
    command_list = []
    s = ""
    f = open(os.path.join(TESTROOT, "triql", file[:-4] + ".triql"))
    f2 = open(os.path.join(TESTROOT, "triql", file))
    output = f2.read().strip().split("\n\n")
    for line in f:
        commands += line
        parts = commands.split(";")
        commands = parts[-1:][0]
        command_list += parts[:-1]
    cmd_out = map(None, command_list, output)
    for comm in cmd_out:
        c.execute(comm[0])
        s = xtuples2str(c.xfetchall())
        if s.strip() != comm[1].strip():
            print_failure()
            print "\nQUERY"
            print("################################################")
            print comm[0].strip()
            print "\nMY OUTPUT"
            print("################################################")
            print s.strip()
            print "\nOUTPUT FROM %s" % file[:-6] + ".out"
            print("################################################")
            print comm[1].strip()
            raise Exception()
        cnx.commit()

def loaddata(cnx):
    print_action("Loading data to database %s" % test_conf.get("global", "dbname"))
    execute_file(cnx, test_conf.get("global", "loaddata"))
    print_success()

def removedata(cnx):
    print_action("Removing data from database %s" % test_conf.get("global", "dbname"))
    execute_file(cnx, test_conf.get("global", "removedata"))
    print_success()

def createas(cnx):
    print_action("Creating tables in database %s" % test_conf.get("global", "dbname"))
    execute_file(cnx, test_conf.get("global", "createas"))
    print_success()

def runtest(cnx, file):
    print_action("Running %s" % file)
    test_file(cnx, file)
    print_success()

def runtests(cnx):
    outfiles = [x for x in os.listdir(os.path.join(TESTROOT, "triql")) if x.endswith(".out")]
    for file in outfiles:
        runtest(cnx, file)

def main(argv=None):
    if argv is None:
        argv = sys.argv

    p = optparse.OptionParser()
    p.add_option("-d", dest="database", help="database name to connect to",
                       default=test_conf.get("global", "dbname"))
    p.add_option("-u", dest="user", help="database user name to connect as")
    p.add_option("-p", dest="password", action="store_true", help="prompt for password")

    (options, args) = p.parse_args()

    options = options.__dict__
    if options["password"] == True:
        try:
            options["password"] = getpass('Enter password: ')
        except EOFError:
            return 0

    if len(argv) == 2 and argv[1].endswith(".triql"):
        outfile = argv[1][:-6] + ".out"
    else:
        outfile = None

    try:
        initdb()
        cnx = triodb.connect(**options)
        loaddata(cnx)
        createas(cnx)
        if outfile is None:
            runtests(cnx)
        else:
            execute_file(cnx, argv[1], open(outfile, "w"))
        removedata(cnx)
        cnx.close()
        print("\n################################################\n")
        if outfile is None:
            print "You passed the test. You are now officially awesome."
        else:
            print "%s generated." % outfile
        print("\n################################################\n")
    except:
        common.print_exception()
        return 1

if __name__ == "__main__":
    sys.exit(main())
