"""Things that are shared among files.
"""

__id__ = "$Id$"

from node import *
import os
import sys
import traceback
import ConfigParser

TYPE_UNCERTAIN = 0
TYPE_CERTAIN = 1
TYPE_CONFIDENCE = 2

SELECTOPTIONS = 0
SELECTCLAUSE = 1
FROMCLAUSE = 2
WHERECLAUSE = 3

GROUPBY = 5
ORDERBY = 6

ROOT = os.path.dirname(os.path.abspath(__file__))

_SQL_FILES = ['conf.sql',
              'extlineage.sql',
              'lavg.sql',
              'havg.sql',
              'eminmax.sql',
              'numg.sql',
              'lavgmain.sql',
              'havgmain.sql',
              'emm.sql',
              'product.sql',
]
SQL_FILES = [os.path.join(ROOT, 'sql', f) for f in _SQL_FILES]

cp = ConfigParser.ConfigParser()
cp.read(os.path.join(ROOT, "trio.cfg"))

def print_exception(msg=''):
    e = sys.exc_info()
    print
    if int(cp.get("global", "debug_level")) > 0:
        print str(e[0]) + ": " + str(e[1])
        traceback.print_tb(e[2])
    else:
        print str(e[1])
    print

def resultTableType(tables, cursor):
    """
    Return the table type of the newly created table based on the
    types of base tables.
    Types:
    0 -> uncertain without confidence
    1 -> certain
    2 -> uncertain with confidence
    """
    types = []
    for table in tables:
        sql = "SELECT type FROM trio_tables WHERE tablename = '%s'" % table
        cursor.execute(sql)
        tup = cursor.fetchall()
        if len(tup) == 0:
            raise Exception("Table '%s' does not exist in Trio schema!" % table)
        elif len(tup) > 1:
            raise Exception("Trio schema inconsistent for table '%s'")
        types.append(tup[0][0])
    if types.count(TYPE_UNCERTAIN) > 0:
        return TYPE_UNCERTAIN
    elif types.count(TYPE_CONFIDENCE) > 0:
        return TYPE_CONFIDENCE
    else:
        return TYPE_CERTAIN

def checkAmbiguity(columnname, columns):
    matched = 0
    for c in columns:
        if (isinstance(c, Dot) and columnname == c.children[1].value) or \
           (isinstance(c, As) and isinstance(c.children[0], Dot) and columnname == c.children[0].children[1].value) or \
           (isinstance(c, As) and isinstance(c.children[0], Node) and columnname == c.children[0].value) or \
           (isinstance(c, As) and isinstance(c.children[0], Node) and columnname == c.children[1].value) or \
           (isinstance(c, Node) and columnname == c.value):
            matched += 1
    return matched

def compare_values(tup1, tup2, datavals):
    for i in range(datavals):
        if tup1[i] < tup2[i]:
            return -1
        elif tup1[i] > tup2[i]:
            return 1
    return 0

def bindparamstr(numparams):
    return ", ".join(["%s" for x in range(numparams)])
