__id__ = "$Id$"

import select
import insert
from trio import common
import createtable
import setoperator
import time
from trio.node import *

def execute(node, trio_cursor, isTemporary=False):
    table = node.children[0].value
    lineage_tables = []
    trio_cursor.isTemporary = isTemporary

    # Run the query and prepare the Trio cursor
    sql = select.execute(node.children[1], trio_cursor, True)

    # Check for proper aliases in expressions
    i = 0
    u_columnnames = []
    columns = isinstance(node.children[1], Select) and node.children[1].children[1].children\
      or node.children[1].children[0].children[1].children
    for child in columns:
        if i < trio_cursor.num_datavalues:
            if not isinstance(child, Dot) and not isinstance(child, As) and len(child.children) > 0:
                raise Exception("Alias required for expressions in CREATE TABLE AS statements!")
            if isinstance(child, Dot) or isinstance(child, As):
                value1 = child.children[1].value
            else:
                value1 = child.value
            j = i + 1
            while j < len(columns):
                if isinstance(columns[j], Dot) or isinstance(columns[j], As):
                    value2 = columns[j].children[1].value
                else:
                    value2 = columns[j].value is not None and columns[j].value or ""
                if value1 == value2:
                    raise Exception("Duplicate or ambiguous column reference '%s'" % value1)
                j += 1
            if i in trio_cursor.u_columns:
                if isinstance(child, Dot) or isinstance(child, As):
                    u_columnnames.append(child.children[1].value)
                else:
                    u_columnnames.append(child.value)
        i += 1

    # Create the new Trio table and forward the select statement to SPI
    createtable.create_trio_table(table, trio_cursor.tableType, trio_cursor.description, u_columnnames, trio_cursor,
                                  trio_cursor.isConfBase, isTemporary)

    start_time = time.time()

    if isinstance(node.children[1], Select):
        lineage_tables = node.children[1].names

        if int(common.cp.get("global", "use_SPI")) == 0:
            # Python-Fetch
            ins = insert.InsertExecutor(table, trio_cursor, uncertainSet=u_columnnames)
            while True:
                xtup = trio_cursor.fetchone()
                if xtup is None:
                    break
                ins.insertXTuple(xtup, trio_cursor)

        else:
            # SPI-Fetch
            conf_option = 0
            if trio_cursor.noconf:
                conf_option = -1
            if trio_cursor.gotAsConf:
                conf_option = 1
            if trio_cursor.gotUniform:
                conf_option = 2
            if not trio_cursor.distinct:
                tables = "|".join("%s" % name for name in lineage_tables) + "|"
            else:
                tables = "%s|" % trio_cursor.distinctTable
            columns = "|".join(x[0] for x in trio_cursor.description) + "|"
            coltypes = "|".join(i in trio_cursor.u_columns and "u" or "c" for i in range(trio_cursor.num_datavalues)) + "|"
            cursor = trio_cursor.tdb.ncursor()
            cursor.execute(("SELECT * FROM trio_selectinto($$%s$$, $$%s$$, $$%s$$, $$%s$$, $$%s$$"
                            + ", %d, %d, %d, %d, %d, %d, %d, %d, %d)")
                           % (table, sql, tables, columns, coltypes, trio_cursor.tableType, isTemporary,
                              trio_cursor.merged, trio_cursor.flatten, trio_cursor.groupAlts,
                              trio_cursor.nolineage, conf_option, trio_cursor.nomaybe, trio_cursor.distinct))
            if cursor.fetchone()[0] == 0:
                raise Exception("SPI call failed!")
            cursor.close()

        createtable.update_trio_sequences(table, trio_cursor)

        # Re-use lineage from temp table for 'DISTINCT' in 'CREATE TABLE AS' statements
        if trio_cursor.distinct and not trio_cursor.nolineage:
            trio_cursor.cursor.execute("INSERT INTO trio_lin_%s (SELECT * FROM trio_lin_%s)"
                                       % (table, trio_cursor.distinctTable))

    elif isinstance(node.children[1], SetOperator):
        # Patch data and lineage for set-operators UNION, INTERSECT, EXCEPT
        setoperator.execute(node.children[1], trio_cursor, table)
        lineage_tables = node.children[1].children[0].names + node.children[1].children[1].names
        createtable.update_trio_sequences(table, trio_cursor)

    end_time = time.time()

    print "SELECT: " + str(max(0, end_time - start_time))

    # Create indexes *after* the bulk inserts
    if not isTemporary:
        createtable.create_trio_indexes(table, trio_cursor.tableType, trio_cursor, isTemporary)

    print "INDEX:  " + str(max(0, end_time - time.time()))

    # Update metadata for schema-level lineage DAG
    if not trio_cursor.nolineage:
        for lin_table in lineage_tables:
            trio_cursor.tdb.pdb.cursor().execute("INSERT INTO trio_tables_lin VALUES ('%s', '%s')"
                                                 % (table, lin_table))

    trio_cursor.isFetchable = False
    return trio_cursor
