__id__ = "$Id$"

from trio import common
from trio.node import *
from trio.visitor import rewriter
from trio.visitor import printer
from trio.visitor import restructurer
from trio.parser import parser

def execute(node, trio_cursor):
    node = rewriter.Rewriter(trio_cursor).visit(node)
    if isinstance(node.children[0], As):
        table = node.children[0].children[0].value
        alias = node.children[0].children[1].value
    else:
        table = node.children[0].value
        alias = table

    fromClause = printer.Printer().visit(node.children[1])
    optWhereClause = printer.Printer().visit(node.children[2])

    _traverseLineage(alias, table, table, fromClause, optWhereClause, trio_cursor, isinstance(node.children[3], Cascade))
    trio_cursor.isFetchable = False

def _traverseLineage(sourceAlias, sourceTable, derivedTable, fromClause, optWhereClause, trio_cursor, cascade=False):
    c = trio_cursor.tdb.pdb.cursor()
    query = ("SELECT DISTINCT trio_tables_lin.tablename "
             + "FROM trio_tables_lin, trio_tables "
             + "WHERE trio_tables_lin.tablename = trio_tables.tablename "
             + "and istemporary = false and srctable = '%s'") % derivedTable
    c.execute(query)
    c.close()

    tables = [tuple[0] for tuple in c.fetchall()]

    # Cascade over derived tables
    if not cascade and len(tables) > 0:
        raise Exception("Cannot drop tuple that has derived tuples!")
    elif cascade and len(tables) > 0:
        for table in tables:
            if table != derivedTable:
                _traverseLineage(sourceAlias, sourceTable, table, fromClause, optWhereClause, trio_cursor, cascade)
                # Clean-up metadata (lineage DAG)
                c = trio_cursor.tdb.pdb.cursor()
                query = ("SELECT DISTINCT trio_tables_lin.srctable "
                         + "FROM trio_tables_lin, trio_tables "
                         + "WHERE trio_tables_lin.tablename = trio_tables.tablename "
                         + "and istemporary = false and trio_tables_lin.tablename = '%s'") % table
                c.execute(query)
                siblings = [tuple[0] for tuple in c.fetchall()]
                for sibling in siblings:
                    c.execute("SELECT count(*) FROM trio_lin_%s WHERE srctable = '%s'"
                              % (table, sibling))
                    n = c.fetchone()
                    if n[0] is None or n[0] == 0:
                        c.execute("DELETE FROM trio_tables_lin WHERE tablename = '%s' and srctable = '%s'"
                                  % (table, sibling))
                c.close()

    # Actual delete after cascade operations
    _deleteTuples(sourceAlias, sourceTable, derivedTable, fromClause, optWhereClause, trio_cursor)

def _deleteTuples(sourceAlias, sourceTable, derivedTable, fromClause, optWhereClause, trio_cursor):
    c = trio_cursor.tdb.pdb.cursor()
    c.execute("SELECT type FROM trio_tables WHERE tablename = '%s'" % derivedTable)
    type = c.fetchone()[0]

    if sourceTable != derivedTable:
        # Get a unique temporary table name
        c.execute("SELECT nextval('trio_tmp')")
        tmp_table = "tmp_delete_" + str(c.fetchone()[0])

        if len(optWhereClause) > 0:
            query = "SELECT *%s,%s as %s_lineage%s and %s_lineage ==> %s" \
                    % (fromClause, derivedTable, derivedTable, optWhereClause, derivedTable, sourceAlias)
        else:
            query = "SELECT *%s,%s as %s_lineage WHERE %s_lineage ==> %s" \
                    % (fromClause, derivedTable, derivedTable, derivedTable, sourceAlias)

        # Rewrite as a new select statement with lineage
        node = parser.parse(query)
        node = rewriter.Rewriter(trio_cursor).visit(node)
        node.children[1] = SelectClause(children=[Dot(children=[Node("distinct %s_lineage"%derivedTable),Node("trio_aid")])])
        node.children[6] = Empty()

        # Create a temporary table with resp. aids and delete data and lineage matching those aids
        c.execute("CREATE TEMPORARY TABLE %s as (%s)" % (tmp_table, printer.Printer().visit(node)))

        # Delete lineage statement
        c.execute("DELETE FROM trio_lin_%s WHERE aid in (SELECT trio_aid FROM %s)"
                  % (derivedTable, tmp_table))

        # Delete data statements
        if type == common.TYPE_CERTAIN:
            c.execute("DELETE FROM trio_c_%s WHERE trio_xid in (SELECT trio_aid FROM %s)"
                      % (derivedTable, tmp_table))
        else:
            c.execute("DELETE FROM trio_u_%s WHERE trio_aid in (SELECT trio_aid FROM %s)"
                      % (derivedTable, tmp_table))
            c.execute("DELETE FROM trio_c_%s WHERE trio_xid not in (SELECT trio_xid FROM trio_u_%s)"
                      % (derivedTable, derivedTable))

    else:
        # Delete lineage statement
        c.execute("DELETE FROM trio_lin_%s WHERE aid in (SELECT %s.trio_aid "
                  % (sourceTable, sourceAlias)
                  + fromClause + optWhereClause + ")")

        # Delete data statements
        if type == common.TYPE_CERTAIN:
            c.execute("DELETE FROM trio_c_%s WHERE trio_xid in (SELECT %s.trio_aid "
                      % (sourceTable, sourceAlias)
                      + fromClause + optWhereClause + ")")
        else:
            c.execute("DELETE FROM trio_u_%s WHERE trio_aid in (SELECT %s.trio_aid "
                      % (sourceTable, sourceAlias)
                      + fromClause + optWhereClause + ")")
            c.execute("DELETE FROM trio_c_%s WHERE trio_xid not in (SELECT trio_xid FROM trio_u_%s)"
                      % (sourceTable, sourceTable))

    c.close()
