__id__ = "$Id$"

import command
import createindex
import createtable
import createtableas
import dropindex
import droptable
import insert
import select
import setoperator

from trio.node import *
from trio.visitor import printer

class Executor:
    def __init__(self, trio_cursor):
        self.tcursor = trio_cursor

    def execute(self, node):
        nodetype = node.__class__.__name__
        executefn = getattr(self, 'execute'+nodetype, self.executeDefault)
        executefn(node)

    def executeDefault(self, node):
        raise Exception("Unsupported command: %s" % node.__class__.__name__)

    def executeCommand(self, node):
        command.CommandExecutor().execute(node, self.tcursor)

    def executeCreateIndex(self, node):
        createindex.execute(node, self.tcursor)

    def executeCreateTable(self, node):
        createtable.execute(node, self.tcursor)

    def executeCreateTableAs(self, node):
        createtableas.execute(node, self.tcursor)

    def executeCreateTempTable(self, node):
        createtable.execute(node, self.tcursor, isConfBase=False, isTemporary=True)

    def executeCreateTempTableAs(self, node):
        createtableas.execute(node, self.tcursor, isTemporary=True)


    def executeDropIndex(self, node):
        dropindex.execute(node, self.tcursor)

    def executeDropTable(self, node):
        droptable.execute(node, self.tcursor)

    def executeInsert(self, node):
        if not isinstance(node.children[1], Empty):
            columnList = [node.children[1].children[i].value for i in range(len(node.children[1].children))]
            insert.InsertExecutor(node.children[0].value, self.tcursor, columnList).execute(node, self.tcursor)
        else:
            insert.InsertExecutor(node.children[0].value, self.tcursor).execute(node, self.tcursor)

    def executeSelect(self, node):
        select.execute(node, self.tcursor)

    def executeSetOperator(self, node):
        setoperator.execute(node, self.tcursor)

