__id__ = "$Id$"

from trio.node import *
from trio import common
import time

# SELECT DISTINCT

# Features: - this algorithm subsumes horizontal merging
#           - merely assumes that at most all alternatives of an xtuple fit in memory at a time
#           - supports on-the-fly confidence computation
# Caution:  - for persistent queries, a second scan is required to get the ?'s at xtuple-level right

# not possible elsewhere because of groupalts
_TRIO_ALTS = -3
_TRIO_ENC = -2
_TRIO_CONF = -1

def execute(node, trio_cursor, sql):
    gotUniform = trio_cursor.gotUniform
    tableType = trio_cursor.tableType
    isConfBase = trio_cursor.isConfBase
    gotAsConf = trio_cursor.gotAsConf
    isTemporary = trio_cursor.isTemporary

    flatten = trio_cursor.flatten
    noconf = trio_cursor.noconf
    nomaybe = trio_cursor.nomaybe
    nolineage = trio_cursor.nolineage
    u_columns = trio_cursor.u_columns

    cursor1 = trio_cursor.tdb.ncursor()
    cursor2 = trio_cursor.tdb.ncursor()

    sqlnodistinct = trio_cursor.sqlCommand[trio_cursor.sqlCommand.lower().find('select'):]
    sqlnodistinct = sqlnodistinct[:sqlnodistinct.lower().find('distinct')] \
                    + sqlnodistinct[sqlnodistinct.lower().find('distinct')+9:]

    columnnames = []
    for i in range(len(trio_cursor.description)):
        columnnames.append(trio_cursor.description[i][0])

    # Get a unique temporary table name
    cursor1.execute("SELECT nextval('trio_temporary')")
    tmp_table = "tmp_distinct_" + str(cursor1.fetchone()[0])

    # Create a temporary table
    trio_cursor.execute("Create Temporary Table %s (%s)"
                        % (tmp_table, ", ".join("%s %s" % (t[0], t[1]) for t in trio_cursor.description)))

    cursor1.execute("UPDATE trio_tables SET confbase = false WHERE tablename = '%s'" % tmp_table)

    # Initialize xid's of input tables
    aid = 1
    xid = 0
    max_xids = []
    i = 0

    # If bit-shifted xid's exceed the available integer range,
    # we may run into numerical problems and need another temp table to materialized the join first
    mayExceedIntRange = False
    for table in node.names:
        cursor1.execute("SELECT nextval('trio_xid_%s')" % table)
        max_xids.append(cursor1.fetchone()[0])
        xid += pow(max_xids[i], i+1)
        if max_xids[i] > 1<<16: # heuristic, actually depends on join-multiplicty too
            mayExceedIntRange = True
            break
        i += 1

    if False and mayExceedIntRange:
        # Create a second temporary table
        trio_cursor.execute("Create Temporary Table %s_tmp As (%s)" % (tmp_table, sqlnodistinct))

        # Move cursor over this temporary table for deduplication
        sql = ("SELECT %s, trio_xid, trio_aid, trio_alts, trio_enc, trio_conf"
               + " FROM %s_tmp "
               + "ORDER BY %s, trio_xid asc"
              ) % (", ".join(x for x in columnnames), tmp_table, ", ".join(x for x in columnnames))

        trio_cursor.num_joins = 1

        node.names = ["%s_tmp" % tmp_table]
        node.aliases = ["%s_tmp" % tmp_table]

        node.aliasdict = {}
        node.aliasdict["%s_tmp" % tmp_table] = "%s_tmp" % tmp_table

        max_xids = []
        cursor1.execute("SELECT nextval('trio_xid_%s_tmp')" % tmp_table)
        max_xids.append(cursor1.fetchone()[0])
        xid = max_xids[0]

    starttime = time.time()

    if int(common.cp.get("global", "use_SPI")) == 1:  # SPI fetch&deduplication

        tables = "|".join("%s" % name for name in node.names) + "|"
        cursor = trio_cursor.tdb.ncursor()
        columns = "|".join(x[0] for x in trio_cursor.description) + "|"
        cursor.execute("SELECT * FROM trio_selectinto_distinct($$%s$$, $$%s$$, $$%s$$, $$%s$$, %d)"
                       % (tmp_table, sql, columns, tables, isinstance(node.children[4], ComputeConfidences)))
        if cursor.fetchone()[0] == 0:
            raise Exception("SPI call failed!")
        cursor.close()

    else: # Python fetch

        # Initialize data & lineage statements
        gotConfStar = isinstance(node.children[4], ComputeConfidences)

        insert_sql = (
            "INSERT INTO " + tmp_table + " ("
            + (", ".join("%s" % t[0] for t in trio_cursor.description))
            + ", trio_xid, trio_aid, trio_alts, trio_enc, trio_conf"
            + ") VALUES ("
            + common.bindparamstr(trio_cursor.num_datavalues)
            + ", %d, %d, %d, %d")
        if gotConfStar:
            insert_sql += ", conf(array['%s']" % tmp_table + ", array[%d], array[1], NULL, 0, 1)"
        else:
            insert_sql += ", %s"
        insert_sql += ")"

        insert_lin_sql = "INSERT INTO trio_lin_%s (aid, disjlin, srclin, srctable) " % tmp_table \
                         + "VALUES (%d, %d, %d, '%s')"

        # Run the query sorted by data columns & xid's
        cursor1.execute(sql)
        tup = cursor1.fetchone()
        if tup is not None:
            questionMark = tup[_TRIO_ALTS] < tup[_TRIO_ENC]

        # Read alts sequentially, first buffers alts with same data and xid
        # - when xid changes, flush lineage into new xtuple & stop buffering, else keep xid
        # - when data changes, flush distinct data with new aid each
        disj = 0
        do_buffer = True
        alts_buffer = []
        while tup != None:
            tmp = tup
            tmp_xid = _getXid(tmp, max_xids, trio_cursor)
            tup = cursor1.fetchone()
            if (tup is None
                or common.compare_values(tmp, tup, trio_cursor.num_datavalues) != 0):
                # flush lineage buffer
                alts_buffer += [tmp]
                for alt in alts_buffer:
                    for lin in _getLin(alt, node.names, trio_cursor):
                        cursor2.execute(insert_lin_sql % (aid, disj, lin[0], lin[1]))
                    disj += 1
                alts_buffer = []
                # insert data
                cols = tmp[:trio_cursor.num_datavalues]
                cols += [int(do_buffer and tmp_xid or xid), aid, \
                         tmp[_TRIO_ALTS], tmp[_TRIO_ALTS] + (questionMark and 1 or 0)]
                if gotConfStar:
                    cols += [aid]
                else:
                    cols += [tmp[_TRIO_CONF]]
                cursor2.execute(insert_sql, cols)
                do_buffer = True
                disj = 0
                # create new aid
                aid += 1
                # update version interval
                if tup is not None:
                    questionMark = tup[_TRIO_ALTS] < tup[_TRIO_ENC]

            else:
                tup_xid = _getXid(tup, max_xids, trio_cursor)
                questionMark = questionMark or tmp[_TRIO_ALTS] < tmp[_TRIO_ENC]
                if do_buffer and tmp_xid == tup_xid:
                    # buffer lineage
                    alts_buffer += [tmp]
                else:
                    if do_buffer:
                        # flush lineage buffer
                        alts_buffer += [tmp]
                        for alt in alts_buffer:
                            for lin in _getLin(alt, node.names, trio_cursor):
                                cursor2.execute(insert_lin_sql % (aid, disj, lin[0], lin[1]))
                            disj += 1
                        alts_buffer = []
                        do_buffer = False
                        # create new xid
                        xid += 1
                    else:
                        # flush lineage w/out buffering
                        for lin in _getLin(tmp, node.names, trio_cursor):
                            cursor2.execute(insert_lin_sql % (aid, disj, lin[0], lin[1]))
                        disj += 1

    cursor1.close()
    cursor2.close()

    endtime = time.time()
    print "DISTINCT: "  + str(endtime - starttime)

    # Initialize SQL cursor on temporary table; this pass also fixes ?'s and reassigns contiguous xid's
    sql = ("SELECT %s, trio_xid, trio_aid, trio_alts, trio_enc, trio_conf"
           + " FROM %s "
           + "ORDER BY trio_xid asc"
          ) % (", ".join(x for x in columnnames), tmp_table)

    trio_cursor.cursor.execute(sql)

    trio_cursor.gotUniform = gotUniform
    trio_cursor.tableType = tableType
    trio_cursor.isConfBase = isConfBase
    trio_cursor.gotAsConf = gotAsConf
    trio_cursor.isTemporary = isTemporary

    trio_cursor.flatten = flatten
    trio_cursor.merged = False
    trio_cursor.noconf = noconf
    trio_cursor.nomaybe = nomaybe
    trio_cursor.nolineage = nolineage

    trio_cursor.distinct = True
    trio_cursor.distinctTable = tmp_table
    trio_cursor.num_joins = 1
    trio_cursor.u_columns = u_columns

    return sql, trio_cursor

def _getLin(tup, tables, trio_cursor):
    lin = []
    for i in range(len(tables)):
        lin.append([tup[trio_cursor.num_datavalues + trio_cursor.num_joins + i], tables[i]])
    return lin

def _getXid(tup, max_xids, trio_cursor):
    xid = 0
    tmp = tup[trio_cursor.num_datavalues:trio_cursor.num_datavalues + trio_cursor.num_joins]
    # may be a numerical problem for large tables or wide joins
    for i in range(len(tmp)):
        xid += tmp[i] * pow(max_xids[i],i)
    return xid
