__id__ = "$Id$"

from trio.node import *
from trio.visitor import rewriter
from trio.visitor import printer
from trio.visitor import dumper
from executor import insert
from trio import common

def execute(node, trio_cursor, deleteOnly=False):
    if trio_cursor.isTemporary:
        raise Exception('Updates not allowed on temporary tables!')

    # Rewrite TriQL to SQL
    r = rewriter.Rewriter(trio_cursor)
    new_node = r.visit(node)

    # Set conf flags, e.g., a derived green table can become a confBase but not have asConf
    trio_cursor.isConfBase = new_node.isConfBase
    trio_cursor.gotAsConf = new_node.gotAsConf
    trio_cursor.gotUniform = new_node.gotUniform

    # Initialize the Trio cursor and determine various fetch options
    trio_cursor.root_node = new_node
    trio_cursor.num_joins = len(new_node.children[2].children)

    table_name = node.children[0].value

    p = printer.Printer(trio_cursor)
    i = insert.InsertExecutor(table_name, trio_cursor)

    sets_certain = False
    sets_uncertain = False
    changes_questionmark = False

    # New ? encoding
    enc = "trio_enc"
    if not isinstance(node.children[2], Empty):
        changes_questionmark = True
        if node.children[2].value == 'addmaybe':
            enc = "trio_alts+1 as trio_enc"
        elif node.children[2].value == 'delmaybe':
            enc = "trio_alts as trio_enc"

    # Certain data columns in SET
    c_cols_set = []
    for c_col in i.c_columnnames:
        matched = False
        if isinstance(node.children[1], Empty):
            for s in node.children[1].children:
                if isinstance(s.children[0], Dot):
                    if s.children[0].children[1].value == c_col:
                        c_cols_set.append(p.visit(s.children[1]) + " as " + c_col)
                        sets_certain = True
                        matched = True
                        break
                else:
                    if s.children[0].value == c_col:
                        c_cols_set.append(p.visit(s.children[1]) + " as " + c_col)
                        sets_certain = True
                        matched = True
                        break
        if not matched:
            c_cols_set.append(c_col)

    # Uncertain data columns in SET
    u_cols_set = []
    u_cols_all = []
    for u_col in i.u_columnnames:
        matched = False
        if isinstance(node.children[1], UpdateSetClause):
            for s in node.children[1].children:
                if isinstance(s.children[0], Dot):
                    if s.children[0].children[1].value == u_col:
                        u_cols_set.append(p.visit(s.children[1]) + " as " + u_col)
                        sets_uncertain = True
                        matched = True
                        break
                else:
                    if s.children[0].value == u_col:
                        u_cols_set.append(p.visit(s.children[1]) + " as " + u_col)
                        sets_uncertain = True
                        matched = True
                        break
        if not matched:
            u_cols_set.append(u_col)
        u_cols_all.append(u_col)

    #print c_cols_set
    #print u_cols_set

    c_cols = []
    u_cols = []
    if trio_cursor.tableType == common.TYPE_CERTAIN:
        c_cols += ["nextval('trio_xid_%s')" % table_name + " as trio_xid"]
        c_cols += ["currval('trio_temporal') as trio_start"]
        c_cols += ["1<<30 as trio_end"]
    else:
        c_cols += ["trio_xid"]
        c_cols += ["trio_alts"]
        c_cols += [enc]
        c_cols += ["currval('trio_temporal') as trio_start"]
        c_cols += ["1<<30 as trio_end"]
        u_cols += ["trio_xid"]
        u_cols += ["nextval('trio_aid_%s')" % table_name + " as trio_aid"]
        u_cols += ["trio_conf"]
        u_cols += ["currval('trio_temporal') as trio_start"]
        u_cols += ["1<<30 as trio_end"]

    c_cols_set += c_cols
    u_cols_set += u_cols
    u_cols_all += u_cols

    # Rewrite WHERE clause for current-version management
    curVersionPredicate = BinaryExpression(value='>', children=[Node('trio_end'), Node("currval('trio_temporal')")])
    expVersionPredicate = BinaryExpression(value='=', children=[Node('trio_end'), Node("currval('trio_temporal')")])

    c = trio_cursor.tdb.pdb.cursor()

    # Expires end-time of existing alternatives
    if not isinstance(node.children[3], Empty):
        predicateSelect = WhereClause(children=[BinaryExpression('and',\
          children=[Parenthesis(children=[curVersionPredicate]),\
          Parenthesis(children=[node.children[3].children[0]])])])
    else:
        predicateSelect = WhereClause(children=[curVersionPredicate])

    if (deleteOnly or sets_certain or changes_questionmark):
        c_update = "UPDATE trio_c_%s SET trio_end = currval('trio_temporal') " % table_name \
                   + "WHERE trio_xid in (SELECT trio_xid FROM " + table_name + " " + p.visit(node.children[3]) + ") " \
                   + "and (" + p.visit(curVersionPredicate) + ")"
        c.execute(c_update)
        #print c_update

        if (sets_uncertain or changes_questionmark):
            view = "SELECT c.trio_xid FROM trio_c_%s c, trio_u_%s u " % (table_name, table_name) \
                   + "WHERE (c.trio_xid = u.trio_xid and c.trio_start = u.trio_start " \
                   + "and c.trio_end = currval('trio_temporal') and u.trio_end > currval('trio_temporal'))"
            if not isinstance(node.children[3], Empty):
                view += " and (" + p.visit(node.children[3].children[0]) + ")"
            u_update = "UPDATE trio_u_%s SET trio_end = currval('trio_temporal') " % table_name \
                       + "WHERE trio_xid in (" + view + ") and (" + p.visit(curVersionPredicate) + ")"
            c.execute(u_update)
            #print u_update

    elif (deleteOnly or sets_uncertain or changes_questionmark):
        u_update = "UPDATE trio_u_%s SET trio_end = currval('trio_temporal')" % table_name
        u_update += p.visit(predicateSelect)
        u_update
        c.execute(u_update)
        #print u_update

    if not deleteOnly:
        # Insert new alternatives for updated xtuples
        if not isinstance(node.children[3], Empty):
            predicateSelect = WhereClause(children=[BinaryExpression('and',\
              children=[Parenthesis(children=[expVersionPredicate]),\
              Parenthesis(children=[node.children[3].children[0]])])])
        else:
            predicateSelect = WhereClause(children=[expVersionPredicate])

        if (sets_certain or changes_questionmark):
            c_insert_set = "SELECT " + ", ".join(c_cols_set) + " "  \
                           + "FROM trio_c_%s " % table_name \
                           + "WHERE trio_xid in (SELECT trio_xid FROM " + table_name + " " \
                           + p.visit(node.children[3]) + ") and (" + p.visit(expVersionPredicate) + ")"
            c.execute(i.insert_c_sql[:i.insert_c_sql.find("VALUES")] + "(" + c_insert_set + ")")
            #print c_insert_set

        if (sets_uncertain or changes_questionmark):
            u_insert_set = "SELECT " + ", ".join(u_cols_set) \
                           + p.visit(FromClause(children=[node.children[0]])) \
                           + p.visit(predicateSelect)
            c.execute(i.insert_u_sql[:i.insert_u_sql.find("VALUES")] + "(" + u_insert_set + ")")
            #print u_insert_set

    if (changes_questionmark):
        u_insert_all = \
            "SELECT " + ", ".join(u_cols_all) \
            + p.visit(FromClause(children=[node.children[0]])) \
            + " WHERE trio_xid in (SELECT trio_xid FROM " + table_name + " " + p.visit(node.children[3]) + ")" \
            + " and trio_aid in (SELECT trio_aid FROM " + table_name + " WHERE trio_end = currval('trio_temporal'))"
        if not isinstance(node.children[3], Empty):
            u_insert_all += " and not (" + p.visit(node.children[3].children[0]) + ")"
        c.execute(i.insert_u_sql[:i.insert_u_sql.find("VALUES")] + "(" + u_insert_all + ")")
        #print u_insert_all

    c.execute("SELECT nextval('trio_temporal')")
    c.close()

    trio_cursor.isXCursor = True
    trio_cursor.isFetchable = False
