#!/usr/bin/env python

__id__ = "$Id$"

import string
from ply import *

keywords = (
    'addmaybe',
    'all',
    'and',
    'any',
    'as',
    'asc',
    'at',
    'avg',
    'by',
    'cascade',
    'compute',
    'conf',
    'confidences',
    'count',
    'create',
    'current',
    'delete',
    'delmaybe',
    'desc',
    'distinct',
    'drop',
    'eavg',
    'ecount',
    'emax',
    'emin',
    'esum',
    'except',
    'exists',
    'extlineage',
    'explain',
    'extraneous',
    'flatten',
    'float',
    'float4',
    'float8',
    'from',
    'group',
    'groupalts',
    'havg',
    'having',
    'hcount',
    'hmax',
    'hmin',
    'hsum',
    'in',
    'index',
    'insert',
    'int',
    'int4',
    'int8',
    'intersect',
    'into',
    'is',
    'like',
    'lineage',
    'lavg',
    'lcount',
    'lmax',
    'lmin',
    'lsum',
    'max',
    'maybe',
    'merged',
    'min',
    'noconf',
    'nolineage',
    'nomaybe',
    'not',
    'null',
    'on',
    'or',
    'order',
    'scaled',
    'scaledbyexp',
    'select',
    'set',
    'sum',
    'table',
    'temporary',
    'text',
    'trio',
    'trio_aid',
    'trio_xid',
    'uncertain',
    'uniform',
    'union',
    'update',
    'values',
    'varchar',
    'version',
    'view',
    'where',
    'with'
)

tokens = keywords + (
    'id',
    'lineage_arrow',
    'comparison_op',
    'string_literal',
    'numeric_literal',
    'command',
)

t_command = r'\\\w'
t_comparison_op = r'!=|<=|>=|<>|<|>|[Ii]?[Ll][Ii][Kk][Ee]'
literals= r'\+-*/.,;()?|[]:='

def t_comment(t):
    r"--.*"
    pass

def t_string_literal(t):
    r"'(\\'|[^'])*'"
    t.value = t.value[1:-1].replace("\\'", "'")
    return t

def t_numeric_literal(t):
    r"\d*\.?\d+([eE][-+]?\d+)?"
    t.value = eval(t.value)
    return t

def t_id(t):
    r"\w+"
    t.value = str(t.value.lower())
    if not t.value.startswith("trio_aid") and not t.value.startswith("trio_xid"):
        if t.value.startswith("trio_"):
            raise Exception('"trio_" prefix is reserved by Trio system.')
    if t.value.startswith("pg_"):
        raise Exception('"pg_" prefix is reserved by Trio system.')
    if t.value.endswith("_pkey"):
        raise Exception('"_pkey" suffix is reserved by Trio system.')
    if t.value in keywords:
        if t.value == 'null':
            t.type = 'null'
            t.value = None
        else:
            t.type = t.value.lower()
    return t

t_ignore = " \t\r"

def t_lineage_arrow(t):
    "==>"
    return t

def t_newline(t):
    r"\n"
    t.lexer.lineno += 1

def t_error(t):
    print "Illegal character", t.value[0]
    t.lexer.skip(1)

lexer=lex.lex(debug=0)
