#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "executor/spi.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

#define MAXDEG 512
#define TABLENAME 256
#define GREENCONF -1
#define MAXTUPS 256
#define MAXALTS 256
#define NOVAL -1

enum linkType { BASE, CON, DIS, NEG };

struct noderec {
    int deg;
    enum linkType type;
    int aid;
    char table[TABLENAME];
    struct noderec *child[MAXDEG];
    int tup;
    int alt;
};

typedef struct noderec node;

static node *child_tree(char *tname, int aid, int aids[MAXTUPS][MAXALTS], int xids[MAXTUPS], double confs[MAXTUPS][MAXALTS],
                        char tables[MAXTUPS][TABLENAME], int *conjunctive, int stopatgreen)
{
    node *res;
    char linquery[8192];
    char metaquery[8192];
    int ret;
    int proc;
    char srctname[MAXDEG][TABLENAME];
    int srcaid[MAXDEG];
    int disj[MAXDEG];
    int j, i;
    int base = 0;

    res = palloc(sizeof(node));
    snprintf(res->table, sizeof(res->table), "%s", tname);
    res->aid = aid;

    //add meta-data query to check whether base

    snprintf(metaquery, sizeof(metaquery), "select confbase, type from trio_tables where tablename='%s'", tname);
    //elog(INFO, "%s", metaquery);

    ret = SPI_execute(metaquery, true, 0);
    proc = SPI_processed;

    if (proc > 1)
        elog(ERROR, "Table Metadata inconsistent!");
    else if (proc == 0)
        base = 0;
    else if (ret > 0 && SPI_tuptable != NULL) {

        TupleDesc tupdesc = SPI_tuptable->tupdesc;
        SPITupleTable *tuptable = SPI_tuptable;
        HeapTuple tuple = tuptable->vals[0];

        if (strcmp(SPI_getvalue(tuple, tupdesc, 1), "t") == 0) {
            base = 1;
        } else {
            base = 0;
        }
        //elog(INFO, "%s %i", tname, base);

        if (stopatgreen == 1) {
            if (atoi(SPI_getvalue(tuple, tupdesc, 2)) == 0)
                base = 1;
        }

    } else {
        elog(ERROR, "Bad SPI result for metaquery");
    }

    if (base == 1) {

        int found = 0;
        res->type = BASE;

        for (j = 0; j < MAXTUPS; j++) {
            for (i = 0; i < MAXALTS; i++) {
                if (aids[j][i] == NOVAL)
                    break;
                if ((aids[j][i] == aid) && (strcmp(tname, tables[j]) == 0)) {
                    res->tup = j;
                    res->alt = i;
                    found = 1;
                }
            }
        }

        if (!found) {

            snprintf(linquery, sizeof(linquery), "select trio_xid,trio_alts,trio_conf,trio_enc from %s where trio_aid=", tname);
            snprintf(linquery + strlen(linquery), sizeof(linquery) - strlen(linquery), "%d", aid);
            //elog(INFO, "DATA-LIN %s\n", linquery);

            ret = SPI_execute(linquery, true, 0);
            proc = SPI_processed;

            if (proc != 1)
                elog(ERROR, "AID not unique or doesn't exist\n");

            if (ret > 0 && SPI_tuptable != NULL) {
                double conf;
                int enc;
                int xid;
                int alts;
                TupleDesc tupdesc = SPI_tuptable->tupdesc;
                SPITupleTable *tuptable = SPI_tuptable;
                HeapTuple tuple = tuptable->vals[0];

                xid = atoi(SPI_getvalue(tuple, tupdesc, 1));
                alts = atoi(SPI_getvalue(tuple, tupdesc, 2));
                if (NULL == SPI_getvalue(tuple, tupdesc, 3))
                    conf = GREENCONF;
                else
                    conf = atof(SPI_getvalue(tuple, tupdesc, 3));
                enc = atoi(SPI_getvalue(tuple, tupdesc, 4));

                if (conf == GREENCONF) {
                    if (alts > enc)
                        conf = 1.0 / (enc + 1);
                    else
                        conf = 1.0 / enc;
                }

                found = 0;
                for (j = 0; j < MAXTUPS; j++) {
                    if (xids[j] == NOVAL) {
                        res->tup = j;
                        res->alt = 0;
                        found = 1;
                        break;
                    }
                    if ((xid == xids[j]) && (strcmp(tname, tables[j]) == 0)) {
                        for (i = 0; i < MAXALTS; i++) {
                            if (aids[j][i] == NOVAL) {
                                res->tup = j;
                                res->alt = i;
                                found = 1;
                                break;
                            }
                        }
                        break;
                    }
                }
                if (!found)
                    elog(ERROR, "MAXTUPS or MAXALTS too small\n");

                xids[res->tup] = xid;
                aids[res->tup][res->alt] = aid;
                confs[res->tup][res->alt] = conf;
                //elog(INFO, " CONF %i %i = %f", res->tup, res->alt, conf);
                snprintf(tables[res->tup], sizeof(tables[res->tup]), "%s", tname);

            } else {
                elog(ERROR, "AID not unique or doesn't exist\n");
            }
        }


    } else {

        snprintf(linquery, sizeof(linquery), "select aid,disjlin,srclin,srctable from trio_lin_%s where aid=", tname);
        snprintf(linquery + strlen(linquery), sizeof(linquery) - strlen(linquery), "%d order by disjlin", aid);
        //elog(INFO, "LIN     %s\n", linquery);

        ret = SPI_execute(linquery, true, 0);
        proc = SPI_processed;

        if (proc > MAXDEG)
            elog(ERROR, "MAXDEG is too small\n");

        if (proc == 0)
            elog(ERROR, " Table type not set properly in trio_tables. Derived table with no lineage found!");

        if (ret > 0 && SPI_tuptable != NULL) {
            TupleDesc tupdesc = SPI_tuptable->tupdesc;
            SPITupleTable *tuptable = SPI_tuptable;
            int lincase = 0;
            int spl = proc;
            for (j = 0; j < proc; j++) {
                HeapTuple tuple = tuptable->vals[j];
                snprintf(srctname[j], sizeof(srctname[j]), "%s", SPI_getvalue(tuple, tupdesc, 4));
                srcaid[j] = atoi(SPI_getvalue(tuple, tupdesc, 3));
                disj[j] = atoi(SPI_getvalue(tuple, tupdesc, 2));

                if (j > 0 && (disj[j] - disj[j - 1]) > lincase) {
                    lincase = disj[j] - disj[j - 1];
                    if (lincase > 1)
                        spl = j;
                    *conjunctive = 0;
                }
            }
            res->tup = NOVAL;
            res->alt = NOVAL;

            /////

            if (lincase > 1) {

                res->deg = 2;
                snprintf(res->table, sizeof(res->table), "%s", tname);

                node *res1;
                node *res2;
                res1 = palloc(sizeof(node));
                res2 = palloc(sizeof(node));
                res1->type = DIS;
                res2->type = DIS;
                res1->tup = NOVAL;
                res2->tup = NOVAL;
                res1->alt = NOVAL;
                res2->alt = NOVAL;
                node *dchild = NULL;

                int lst;

                lst = -1;
                res1->deg = 0;
                for (j = 0; j < spl; j++) {
                    if (disj[j] != lst) {
                        dchild = palloc(sizeof(node));
                        res1->child[res1->deg] = dchild;
                        res1->deg++;
                        lst = disj[j];
                        dchild->aid = disj[j];
                        snprintf(res1->table, sizeof(res1->table), "%s", tname);
                        dchild->tup = NOVAL;
                        dchild->alt = NOVAL;
                        dchild->deg = 0;
                        dchild->type = CON;
                    }
                    dchild->deg++;
                }

                lst = -1;
                res2->deg = 0;
                for (j = spl; j < proc; j++) {
                    if (disj[j] != lst) {
                        dchild = palloc(sizeof(node));
                        res2->child[res2->deg] = dchild;
                        res2->deg++;
                        lst = disj[j];
                        dchild->aid = disj[j];
                        snprintf(res2->table, sizeof(res2->table), "%s", tname);
                        dchild->tup = NOVAL;
                        dchild->alt = NOVAL;
                        dchild->deg = 0;
                        dchild->type = CON;
                    }
                    dchild->deg++;
                }

                int l;
                int k;
                int ct = 0;

                for (l = 0; l < res1->deg; l++) {
                    for (k = 0; k < res1->child[l]->deg; k++) {
                        res1->child[l]->child[k] =
                            child_tree(srctname[ct], srcaid[ct], aids, xids, confs, tables, conjunctive, stopatgreen);
                        ct++;
                    }
                }

                for (l = 0; l < res2->deg; l++) {
                    for (k = 0; k < res2->child[l]->deg; k++) {
                        res2->child[l]->child[k] =
                            child_tree(srctname[ct], srcaid[ct], aids, xids, confs, tables, conjunctive, stopatgreen);
                        ct++;
                    }
                }

                res->child[0] = res1;
                res->type = CON;
                if (lincase == 2) {
                    res->child[1] = res2;
                } else if (lincase == 3) {
                    node *negnode;
                    negnode = palloc(sizeof(node));
                    negnode->deg = 1;
                    negnode->type = NEG;
                    negnode->tup = NOVAL;
                    negnode->alt = NOVAL;
                    negnode->child[0] = res2;
                    snprintf(negnode->table, sizeof(negnode->table), "%s", tname);
                    res->child[1] = negnode;
                }

            } else if (lincase == 1) {

                int lst = -1;
                res->type = DIS;
                res->deg = 0;
                node *dchild = NULL;
                for (j = 0; j < proc; j++) {
                    if (disj[j] != lst) {
                        dchild = palloc(sizeof(node));
                        res->child[res->deg] = dchild;
                        res->deg++;
                        lst = disj[j];
                        dchild->aid = disj[j];
                        snprintf(res->table, sizeof(res->table), "%s", tname);
                        dchild->tup = NOVAL;
                        dchild->alt = NOVAL;
                        dchild->deg = 0;
                        dchild->type = CON;
                    }
                    dchild->deg++;
                }

                int l;
                int k;
                int ct = 0;

                for (l = 0; l < res->deg; l++) {
                    for (k = 0; k < res->child[l]->deg; k++) {
                        res->child[l]->child[k] =
                            child_tree(srctname[ct], srcaid[ct], aids, xids, confs, tables, conjunctive, stopatgreen);
                        ct++;
                    }
                }

            } else if (lincase == 0) {

                res->type = CON;
                res->tup = NOVAL;
                res->alt = NOVAL;
                res->deg = proc;
                for (j = 0; j < proc; j++) {
                    res->child[j] = child_tree(srctname[j], srcaid[j], aids, xids, confs, tables, conjunctive, stopatgreen);
                }

            }
        } else {
            elog(ERROR, "Unexpected SPI result");
        }
    }
    return res;
}

static int eval(node *tree, int choice[MAXTUPS])
{
    int i;

    if (tree->type == BASE)
        return choice[tree->tup] == tree->alt;

    if (tree->type == DIS) {
        for (i = 0; i < tree->deg; i++) {
            if (eval(tree->child[i], choice))
                return 1;
        }
        return 0;
    }

    if (tree->type == CON) {
        for (i = 0; i < tree->deg; i++) {
            if (eval(tree->child[i], choice) == 0)
                return 0;
        }
        return 1;
    }

    if (tree->type == NEG) {
        if (eval(tree->child[0], choice) == 0)
            return 1;
        else
            return 0;
    }

    elog(ERROR, "Bad Type");
    return 0;
}

static double montecarlo(int level, double confs[MAXTUPS][MAXALTS], int choice[MAXTUPS], node *root, int *maxLevel, double pw)
{
    int i, j, alts, N = 16000;
    double ret = 0.0, res;

    for (i = 0; i < N; i++) {
        for (j = 0; j < MAXTUPS && confs[j][0] > NOVAL; j++) {
            alts = 0;
            while (confs[j][alts] > NOVAL && alts < MAXALTS)
                alts++;
            choice[j] = rand() % alts;
        }
        if (eval(root, choice)) {
            res = 1.0;
            for (j = 0; j < MAXTUPS; j++) {
                if (choice[j] != NOVAL)
                    res = res * confs[j][choice[j]];
            }
            ret += res;
        }
    }

    res = (ret * pw) / N;
    if (res > 1.0)
        return 1.0;
    return res;
}

static double reciter(int level, double confs[MAXTUPS][MAXALTS], int choice[MAXTUPS], node *root, int *maxLevel)
{
    int i = 0, j;
    double ret = 0.0;

    while (confs[level][i] > NOVAL && i < MAXALTS) {
        choice[level] = i;
        if (level > *maxLevel)
            (*maxLevel) = level;

        if (level + 1 < MAXTUPS && confs[level + 1][0] > NOVAL && level < 12) {
            ret += reciter(level + 1, confs, choice, root, maxLevel);

        } else {

            if (eval(root, choice)) {
                double res = 1.0;
                for (j = 0; j < MAXTUPS; j++) {
                    if (choice[j] != NOVAL)
                        res = res * confs[j][choice[j]];
                }
                ret += res;
            }
        }
        i++;
    }
    return ret;
}

Datum trio_conf_spi(text *p_tables, text *p_aids, text *p_disj, int stopatgreen)
{
    double res;
    node *root;
    int aids[MAXTUPS][MAXALTS];
    int xids[MAXTUPS];
    int choice[MAXTUPS];
    double confs[MAXTUPS][MAXALTS];
    char tables[MAXTUPS][TABLENAME];
    int j, i;
    int *conjunctive = palloc(sizeof(int));

    int proc;
    char argtname[MAXDEG][TABLENAME];
    int argaid[MAXDEG];
    int disj[MAXDEG];
    int topcount;

    *conjunctive = 1;

    char *t_tables = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(p_tables)));
    char *t_aids = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(p_aids)));
    char *t_disj = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(p_disj)));

    char *tok = strtok(t_tables, "|");
    j = 0;
    while (tok != NULL) {
        strcpy(argtname[j], tok);
        tok = strtok(NULL, "|");
        j++;
    }
    proc = j;
    tok = strtok(t_aids, "|");
    j = 0;
    while (tok != NULL) {
        argaid[j] = atoi(tok);
        tok = strtok(NULL, "|");
        j++;
    }
    tok = strtok(t_disj, "|");
    j = 0;
    topcount = 1;
    while (tok != NULL) {
        disj[j] = atoi(tok);
        if ((j > 0) && (disj[j] != disj[j - 1]))
            topcount++;
        tok = strtok(NULL, "|");
        j++;
    }
    root = palloc(sizeof(node));
    for (j = 0; j < MAXTUPS; j++) {
        xids[j] = NOVAL;
        choice[j] = NOVAL;
        for (i = 0; i < MAXALTS; i++) {
            aids[j][i] = NOVAL;
            confs[j][i] = NOVAL;
        }
    }

    SPI_connect();
    if (topcount != 1) {
        root->type = DIS;
        *conjunctive = 0;
        int lst = -1;
        root->deg = 0;
        node *dchild = NULL;
        for (j = 0; j < proc; j++) {
            if (disj[j] != lst) {
                dchild = palloc(sizeof(node));
                root->child[root->deg] = dchild;
                root->deg++;
                lst = disj[j];
                dchild->aid = disj[j];
                snprintf(root->table, sizeof(root->table), "onthefly");
                dchild->tup = NOVAL;
                dchild->alt = NOVAL;
                dchild->deg = 0;
                dchild->type = CON;
            }
            dchild->deg++;
        }

        int l;
        int k;
        int ct = 0;
        for (l = 0; l < root->deg; l++) {
            for (k = 0; k < root->child[l]->deg; k++) {
                root->child[l]->child[k] =
                    child_tree(argtname[ct], argaid[ct], aids, xids, confs, tables, conjunctive, stopatgreen);
                ct++;
            }
        }

    } else {

        root->type = CON;
        root->tup = NOVAL;
        root->alt = NOVAL;
        root->deg = proc;
        for (j = 0; j < proc; j++) {
            root->child[j] = child_tree(argtname[j], argaid[j], aids, xids, confs, tables, conjunctive, stopatgreen);
        }
    }

    if (*conjunctive == 0) {

        double pw = 1.0;
        for (j = 0; j < MAXTUPS; j++) {
            double confsum = 0;
            int last = NOVAL;
            for (i = 0; i < MAXALTS; i++) {
                if (confs[j][i] == NOVAL) {
                    last = i;
                    break;
                } else {
                    confsum += confs[j][i];
                }
            }
            pw *= last;
            if ((confsum < 1) && (last > 0))
                confs[j][last] = 1 - confsum;
        }

        int maxLevel = 0;
        if (pw < 16000)                // exact confs OK until 16000 PWs
            res = reciter(0, confs, choice, root, &maxLevel);
        else {
            //elog(WARNING, "WARNING: maximum recursion level %d for conf() exceeded! Using Monte Carlo simulations instead.", maxLevel);
            res = montecarlo(0, confs, choice, root, &maxLevel, pw);
        }

    } else {

        res = 1;
        for (i = 0; i < MAXTUPS; i++) {
            if (confs[i][0] == NOVAL) {
                break;
            }
            if (confs[i][1] != NOVAL) {
                res = 0;
                break;
            }
            res = res * confs[i][0];
        }

    }

    // Clean up
    SPI_finish();

    PG_RETURN_FLOAT8(res);
}
