/*
 * Return a confidence value of a set of alternatives. This is the 
 * implementation of Trio's CONF() built-in function.
 *
 * $Id$
 */

/*
 * Return the confidence value of a set of alternatives, which is a product
 * of confidences of the alternatives. The confidence value is zero if 2 or
 * more alternatives belong to a same xtuple. If a same alternative appears
 * more than once, its confidence value gets multiplied only once. 
 *
 * This is the top-level function that gets called by the trio system.
 *
 * PRECONDITION:
 *     All the tables must be either of type "with conf" or "certain".
 *
 * @param $1 - table names
 * @param $2 - aids
 * @param $3 - disj-lineage value (same as aids if only conjunctive)
 * @param $4 - stop at green tables (0/1)
 */


CREATE OR REPLACE FUNCTION trio_conf_spi(text, text, text, int) RETURNS float8
    AS 'triospi'
    LANGUAGE C STRICT;

CREATE OR REPLACE FUNCTION conf(text[], integer[], integer[], float8[], int, int) RETURNS float8 AS
$$
DECLARE
    result float8;
    tables text;
    aids text;
    disjlin text;
BEGIN
    if array_upper($1, 1) = 1 and $4[1] is not null then
        return $4[1];
    end if;
    tables = '';
    aids = '';
    disjlin = '';
    for i in 1..array_upper($1, 1) loop
        tables = tables||$1[i]||'|';
        aids = aids||$2[i]||'|';
        disjlin = disjlin||$3[i]||'|';
    end loop;
    select into result * from trio_conf_spi(tables, aids, disjlin, $5);
    if array_upper($1, 1) = 1 and $6 = 1 then
        execute 'update '|| $1[1] ||' set trio_conf=' || result || ' where trio_aid=' || $2[1];
    end if;    
    return result;
END;
$$ LANGUAGE plpgsql;

