/*
 * Create Trio system tables. 
 *
 * trio_attribs
 *  Contains attribute information about user defined tables.  
 *     - tablename: Table name
 *     - attrname : Attribute name
 *     - colnum   : Column index number (starts from 0)
 *
 * trio_tables_lin
 *  Contains table level lineage information.
 *     - tablename: Derived table name
 *     - srctable : Source table name
 * 
 * trio_tables
 *  Contains information about user defined tables. 
 *     - tablename: Table name
 *     - type     : Table type. Possible values are:
 *                      0 for uncertain table
 *                      1 for certain table
 *                      2 for uncertain table with confidence values. 
 *                  These values are defined in common.py.
 * 
 * $Id$
 */
 
/*
drop table trio_attribs;
drop table trio_tables_lin;
drop table trio_tables;
*/

create table trio_attribs(tablename name, attrname name, colnum int, iscertain boolean);
create index trio_attribs_idx on trio_attribs(tablename, attrname, colnum);

create table trio_tables_lin(tablename name, srctable name);
create index trio_tables_lin_idx on trio_tables_lin(tablename, srctable);

create table trio_tables(tablename name, type int, confbase boolean, istemporary boolean);
create index trio_tables_idx on trio_tables(tablename, type, confbase);

-- create sequences for vertical aggregations
create sequence aggs_trio_aid;
create sequence aggs_trio_xid;

-- create sequence for multi-session temporary table names
create sequence trio_temporary;

-- create sequence for multi-session temporal intervals
create sequence trio_temporal start with 2;

