__id__ = "$Id$"

from visitor import Visitor
from trio.node import *
from trio.common import *
import copy

class AggRewriter(Visitor):
    def __init__(self, trio_cursor):
        self.trio_cursor = trio_cursor
        self.stack = []

    def visit(self, node):
        self.stack.append(node)
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        new_node = visitfn(node)
        self.stack.pop(-1)
        return new_node

    def visitDefault(self, node):
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def getAggAlias(self, default):
        try:
            aliasStr = getattr(self, 'aggalias')
        except AttributeError:
            aliasStr = default
        return aliasStr

    def attachAggAlias(self, node):
        lower = node.__class__.__name__.lower()
        aliasStr = self.getAggAlias(lower)
        asNode = As(children=[node, Node(aliasStr)])
        return asNode

    def v2SelectClause(self, node):
        self.stack.append(node)
        if 'hmin' == self.stack[-2].aggType:
            new_children = []

            minnode = Min(children=[Node('certainMaxattr')])
            asChmin = As(children=[minnode,Node('certainHmin')])

            maxnode = Max(children=[Node('uncertainMaxattr')])
            asUhmin = As(children=[maxnode,Node('uncertainHmin')])

            new_children.append(asChmin)
            new_children.append(asUhmin)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
        if 'lmax' == self.stack[-2].aggType:
            new_children = []

            maxnode = Max(children=[Node('certainMinattr')])
            asClmax = As(children=[maxnode,Node('certainLmax')])

            minnode = Min(children=[Node('uncertainMinattr')])
            asUlmax = As(children=[minnode,Node('uncertainLmax')])

            new_children.append(asClmax)
            new_children.append(asUlmax)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
        self.stack.pop(-1)
        return selectclause
    def visitSelectClause(self, node):
        if 'lmax' == self.stack[-2].aggType:
            new_children = []

            casestr = '(case when certainLmax is null then uncertainLmax else certainLmax end)'

            aliasStr = self.getAggAlias('lmax')

            aslmax = As(children=[Node(casestr),Node(aliasStr)])

            new_children.append(aslmax)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
            return selectclause

        if 'lsum' == self.stack[-2].aggType:
            new_children = []
            sumnode = Lsum(children=[Node('minattr')])
            aliasStr = self.getAggAlias('lsum')
            aslsum = As(children=[sumnode,Node(aliasStr)])
            new_children.append(aslsum)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
            return selectclause
        if 'hsum' == self.stack[-2].aggType:
            new_children = []
            sumnode = Hsum(children=[Node('maxattr')])
            aliasStr = self.getAggAlias('hsum')
            ashsum = As(children=[sumnode,Node(aliasStr)])
            new_children.append(ashsum)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
            return selectclause
        if 'hmin' == self.stack[-2].aggType:
            new_children = []

            casestr = '(case when certainHmin is null then uncertainHmin else certainHmin end)'

            aliasStr = self.getAggAlias('hmin')

            ashmin = As(children=[Node(casestr),Node(aliasStr)])

            new_children.append(ashmin)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
            return selectclause
        if 'esum' == self.stack[-2].aggType:
            new_children = []
            numerator = Sum(children=[Node('weighted')])
            predenom= BinaryExpression('-', [Node('1'), Node('product(1-qconf)')])
            denominator = Parenthesis(children=[predenom])
            quotient = BinaryExpression('/', [numerator, denominator])

            aliasStr = self.getAggAlias('esum')
            asesum = As(children=[quotient,Node(aliasStr)])

            new_children.append(asesum)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
            return selectclause
        if 'emin' == self.stack[-2].aggType:
            new_children = []
            numerator = Sum(children=[Node('weighted')])
            predenom= BinaryExpression('-', [Node('1'), Node('product(1-qconf)')])
            denominator = Parenthesis(children=[predenom])
            quotient = BinaryExpression('/', [numerator, denominator])
            asesum = As(children=[quotient,Node('esum')])
            new_children.append(asesum)
            for child in node.children:
                if not isinstance(child, AggInternal):
                    new_children.append(child)
            selectclause = SelectClause(children=new_children)
            return selectclause
        return node

    def visitGroupByKey(self, node):
        new_children = []
        for child in node.children:
            if child.value != 'trio_xid':
                new_children.append(child)
        groupbyclause = GroupByKey(children=new_children)
        return groupbyclause

    def visitGBC(self, node):
        new_children = []
        for child in node.children:
            if child.value not in ['trio_xid', 'trio_enc', 'trio_alts']:
                new_children.append(child)
        groupbyclause = GroupByKey(children=new_children)
        return groupbyclause

    def visitHavgQ1SC(self, node):
        new_children = []
        sumnode = Sum(children=[Node('maxattr')])
        asmsum = As(children=[sumnode,Node('msum')])
        new_children.append(asmsum)
        countnode = Count(children=[Node('*')])
        asmcount = As(children=[countnode,Node('mcount')])
        new_children.append(asmcount)
        for child in node.children:
            if not isinstance(child, AggInternal):
                new_children.append(child)
        selectclause = SelectClause(children=new_children)
        return selectclause

    def visitHavgQ1(self, node):
        select = Select(children=[]);
        select.isConfBase = node.isConfBase
        select.gotAsConf = node.gotAsConf
        select.gotUniform = node.gotUniform
        select.u_columns = node.u_columns
        select.names = node.names
        select.gotAgg = node.gotAgg
        select.aggType = node.aggType

        select.children.append(SelectOptions(children=[Empty()]))

        retselectclause = self.visitHavgQ1SC(node.children[SELECTCLAUSE])
        select.children.append(retselectclause)

        #from clause
        paren = Parenthesis(children=[node])
        subquery = As(children=[paren, Node('s1')])
        select.children.append(FromClause(children=[subquery]))

        select.children.append(Empty())#3
        select.children.append(Empty())#4

        retgroupby = self.visit(node.children[GROUPBY])

        select.children.append(retgroupby)

        select.children.append(Empty())
        select.children.append(Empty())
        select.children.append(Empty())

        return select

    def visitHavgG1SC(self, node):
        new_children = []
        sumnode = Sum(children=[Node('maxattr')])
        asmsum = As(children=[sumnode,Node('msum')])
        new_children.append(asmsum)
        countnode = Count(children=[Node('*')])
        asmcount = As(children=[countnode,Node('mcount')])
        new_children.append(asmcount)

        selectclause = SelectClause(children=new_children)
        selectclause.children.append(Node('g'))
        return selectclause

    def visitHavgG1(self, node):
        select = Select(children=[]);
        select.isConfBase = node.isConfBase
        select.gotAsConf = node.gotAsConf
        select.gotUniform = node.gotUniform
        select.u_columns = node.u_columns
        select.names = node.names
        select.gotAgg = node.gotAgg
        select.aggType = node.aggType

        select.aggAttr = node.aggAttr

        select.children.append(SelectOptions(children=[Empty()]))

        retselectclause = self.visitHavgG1SC(node.children[SELECTCLAUSE])
        select.children.append(retselectclause)

        #from clause
        paren = Parenthesis(children=[node])
        subquery = As(children=[paren, Node('s1')])
        select.children.append(FromClause(children=[subquery]))

        select.children.append(Empty())#3
        select.children.append(Empty())#4
        select.children.append(GroupByKey(children=[Node('g')]))#5

        select.children.append(Empty())
        select.children.append(Empty())
        select.children.append(Empty())

        return select

    def visitLavgQ1SC(self, node):
        new_children = []
        sumnode = Sum(children=[Node('minattr')])
        asmsum = As(children=[sumnode,Node('msum')])
        new_children.append(asmsum)
        countnode = Count(children=[Node('*')])
        asmcount = As(children=[countnode,Node('mcount')])
        new_children.append(asmcount)
        for child in node.children:
            if not isinstance(child, AggInternal):
                new_children.append(child)
        selectclause = SelectClause(children=new_children)
        return selectclause

    def visitLavgQ1(self, node):
        select = Select(children=[]);
        select.isConfBase = node.isConfBase
        select.gotAsConf = node.gotAsConf
        select.gotUniform = node.gotUniform
        select.u_columns = node.u_columns
        select.names = node.names
        select.gotAgg = node.gotAgg
        select.aggType = node.aggType

        select.children.append(SelectOptions(children=[Empty()]))

        retselectclause = self.visitLavgQ1SC(node.children[SELECTCLAUSE])
        select.children.append(retselectclause)

        #from clause
        paren = Parenthesis(children=[node])
        subquery = As(children=[paren, Node('s1')])
        select.children.append(FromClause(children=[subquery]))

        select.children.append(Empty())#3
        select.children.append(Empty())#4

        retgroupby = self.visit(node.children[GROUPBY])

        select.children.append(retgroupby)

        select.children.append(Empty())
        select.children.append(Empty())
        select.children.append(Empty())

        return select

    def visitLavgG1SC(self, node):
        new_children = []
        sumnode = Sum(children=[Node('minattr')])
        asmsum = As(children=[sumnode,Node('msum')])
        new_children.append(asmsum)
        countnode = Count(children=[Node('*')])
        asmcount = As(children=[countnode,Node('mcount')])
        new_children.append(asmcount)

        selectclause = SelectClause(children=new_children)
        selectclause.children.append(Node('g'))
        return selectclause

    def visitLavgG1(self, node):
        select = Select(children=[]);
        select.isConfBase = node.isConfBase
        select.gotAsConf = node.gotAsConf
        select.gotUniform = node.gotUniform
        select.u_columns = node.u_columns
        select.names = node.names
        select.gotAgg = node.gotAgg
        select.aggType = node.aggType

        select.aggAttr = node.aggAttr

        select.children.append(SelectOptions(children=[Empty()]))

        retselectclause = self.visitLavgG1SC(node.children[SELECTCLAUSE])
        select.children.append(retselectclause)

        #from clause
        paren = Parenthesis(children=[node])
        subquery = As(children=[paren, Node('s1')])
        select.children.append(FromClause(children=[subquery]))

        select.children.append(Empty())#3
        select.children.append(Empty())#4
        select.children.append(GroupByKey(children=[Node('g')]))#5

        select.children.append(Empty())
        select.children.append(Empty())
        select.children.append(Empty())

        return select

    def visitSelect(self, node):
        if node.gotAgg != 1:
            return node
        if node.aggType in []:
            return node
        if node.aggType in ['lsum', 'hsum']:
            select = Select(children=[]);
            select.isConfBase = node.isConfBase
            select.gotAsConf = node.gotAsConf
            select.gotUniform = node.gotUniform
            select.u_columns = node.u_columns
            select.names = node.names
            select.gotAgg = node.gotAgg
            select.aggType = node.aggType

            select.children.append(SelectOptions(children=[Empty()]))

            retselectclause = self.visit(node.children[SELECTCLAUSE])
            select.children.append(retselectclause)

            #from clause
            paren = Parenthesis(children=[node])
            subquery = As(children=[paren, Node('s1')])
            select.children.append(FromClause(children=[subquery]))

            select.children.append(Empty())#3
            select.children.append(Empty())#4

            if node.aggType in ['lsum', 'hsum']:
                retgroupby = self.visitGBC(node.children[GROUPBY])
            else:
                retgroupby = self.visit(node.children[GROUPBY])

            select.children.append(retgroupby)

            select.children.append(Empty())
            select.children.append(Empty())
            select.children.append(Empty())

            return select
        if node.aggType in ['lmax', 'hmin']:
            select = Select(children=[]);
            select.isConfBase = node.isConfBase
            select.gotAsConf = node.gotAsConf
            select.gotUniform = node.gotUniform
            select.u_columns = node.u_columns
            select.names = node.names
            select.gotAgg = node.gotAgg
            select.aggType = node.aggType

            select.children.append(SelectOptions(children=[Empty()]))

            retselectclause = self.visit(node.children[SELECTCLAUSE])
            select.children.append(retselectclause)


            select2 = Select(children=[])
            select2.children.append(SelectOptions(children=[Empty()]))
            ret2selectclause = self.v2SelectClause(node.children[SELECTCLAUSE])
            select2.children.append(ret2selectclause)

            #from clause
            paren = Parenthesis(children=[node])
            subquery = As(children=[paren, Node('s1')])
            select2.children.append(FromClause(children=[subquery]))

            select2.children.append(Empty())#3
            select2.children.append(Empty())#4
            retgroupby = self.visit(node.children[GROUPBY])

            select2.children.append(retgroupby)

            select2.children.append(Empty())
            select2.children.append(Empty())
            select2.children.append(Empty())


            #from clause
            paren2 = Parenthesis(children=[select2])
            subquery2 = As(children=[paren2, Node('s2')])
            select.children.append(FromClause(children=[subquery2]))

            select.children.append(Empty())#3
            select.children.append(Empty())#4
            select.children.append(Empty())#5


            select.children.append(Empty())
            select.children.append(Empty())
            select.children.append(Empty())

            return select

        if 'esum' == node.aggType:
            select = Select(children=[]);
            select.isConfBase = node.isConfBase
            select.gotAsConf = node.gotAsConf
            select.gotUniform = node.gotUniform
            select.u_columns = node.u_columns
            select.names = node.names
            select.gotAgg = node.gotAgg
            select.aggType = node.aggType

            select.children.append(SelectOptions(children=[Empty()]))

            retselectclause = self.visit(node.children[SELECTCLAUSE])
            select.children.append(retselectclause)

            #from clause
            paren = Parenthesis(children=[node])
            subquery = As(children=[paren, Node('s1')])
            select.children.append(FromClause(children=[subquery]))

            select.children.append(Empty())#3
            select.children.append(Empty())#4

            retgroupby = self.visit(node.children[GROUPBY])
            select.children.append(retgroupby)

            select.children.append(Empty())
            select.children.append(Empty())
            select.children.append(Empty())

            return select

        if 'havg' == node.aggType:
            select = Select(children=[])
            select.isConfBase = node.isConfBase
            select.gotAsConf = node.gotAsConf
            select.gotUniform = node.gotUniform
            select.u_columns = node.u_columns
            select.names = node.names
            select.gotAgg = node.gotAgg
            select.aggType = node.aggType

            select.children.append(SelectOptions(children=[Empty()]))

            if node.groupByAgg:
                new_children = []
                gbattrs = []
                asgbs = []
                orderbys = []
                i=1
                for child in node.children[SELECTCLAUSE].children:
                    if isinstance(child, AggInternal):
                        new_children.append(child)
                    else:
                        gbchild = child
                        gbname = gbchild.value
                        gbattr = GBAttr(gbname)
                        gbattrs.append(gbattr)
                        asgb = As(children=[Node('temp.g['+str(i)+']'), Node(gbname)])
                        orderby = OrderById(children=[Node(gbname)])
                        orderbys.append(orderby)
                        i+=1
                        asgbs.append(asgb)
                node.children[SELECTCLAUSE] = SelectClause(children=new_children)
                select.aggAttr = node.aggAttr

                aggAttr = node.aggAttr

                gbarray = GBArray(children=gbattrs)
                node.children[SELECTCLAUSE].children.append(gbarray)

                child1 = copy.deepcopy(node)
                child2 = copy.deepcopy(node)
                certaincond = BinaryExpression(value='=', children=[Node('trio_enc'), Node('trio_alts')])

                if isinstance(child1.children[WHERECLAUSE], Empty):
                    child1.children[WHERECLAUSE] = WhereClause(children=[certaincond])
                else:
                    child1.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                      children=[Parenthesis(children=[certaincond]),\
                                Parenthesis(children=child1.children[WHERECLAUSE].children)])])

                notcertaincond = BinaryExpression(value='!=', children=[Node('trio_enc'), Node('trio_alts')])

                if isinstance(child2.children[WHERECLAUSE], Empty):
                    child2.children[WHERECLAUSE] = WhereClause(children=[notcertaincond])
                else:
                    child2.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                      children=[Parenthesis(children=[notcertaincond]),\
                                Parenthesis(children=child2.children[WHERECLAUSE].children)])])

                orderid = OrderById(children=[Node('maxattr desc')])
                orderbys.append(orderid)
                child2.children[ORDERBY] = OrderBy(children=orderbys)

                retchild1 = self.visitHavgG1(child1)
                retchild2 = child2

                havg = Havgmain(children=[retchild1, retchild2])

                asHavg = As(children=[havg, Node('temp')])

                select.children.append(SelectClause(children=[]))

                aliasStr = self.getAggAlias('havg')

                select.children[SELECTCLAUSE].children.append(Node('temp.a as ' + aliasStr))
                for asgb in asgbs:
                    select.children[SELECTCLAUSE].children.append(asgb)

                select.children.append(FromClause(children=[asHavg]))#2
                select.children.append(Empty())#3
                select.children.append(Empty())#4
                select.children.append(Empty())#5
                select.children.append(Empty())#6
                select.children.append(Empty())#7
                select.children.append(Empty())#8

                return select

            child1 = copy.deepcopy(node)
            child2 = copy.deepcopy(node)
            certaincond = BinaryExpression(value='=', children=[Node('trio_enc'), Node('trio_alts')])

            if isinstance(child1.children[WHERECLAUSE], Empty):
                child1.children[WHERECLAUSE] = WhereClause(children=[certaincond])
            else:
                child1.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                  children=[Parenthesis(children=[certaincond]),\
                            Parenthesis(children=child1.children[WHERECLAUSE].children)])])

            notcertaincond = BinaryExpression(value='!=', children=[Node('trio_enc'), Node('trio_alts')])

            if isinstance(child2.children[WHERECLAUSE], Empty):
                child2.children[WHERECLAUSE] = WhereClause(children=[notcertaincond])
            else:
                child2.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                  children=[Parenthesis(children=[notcertaincond]),\
                            Parenthesis(children=child2.children[WHERECLAUSE].children)])])

            orderid = OrderById(children=[Node('maxattr')])
            child2.children[ORDERBY] = OrderByDesc(children=[orderid])

            retchild1 = self.visitHavgQ1(child1)
            retchild2 = child2

            havg = Havg(children=[retchild1, retchild2])

            asHavg = self.attachAggAlias(havg)

            select.children.append(SelectClause(children=[asHavg]))

            select.children.append(Empty())#2
            select.children.append(Empty())#3
            select.children.append(Empty())#4

            retgroupby = self.visit(node.children[GROUPBY])

            select.children.append(retgroupby)

            select.children.append(Empty())#6
            select.children.append(Empty())#7
            select.children.append(Empty())#8

            return select

        if 'lavg' == node.aggType:
            select = Select(children=[])
            select.isConfBase = node.isConfBase
            select.gotAsConf = node.gotAsConf
            select.gotUniform = node.gotUniform
            select.u_columns = node.u_columns
            select.names = node.names
            select.gotAgg = node.gotAgg
            select.aggType = node.aggType

            select.children.append(SelectOptions(children=[Empty()]))

            if node.groupByAgg:
                new_children = []
                gbattrs = []
                asgbs = []
                orderbys = []
                i=1
                for child in node.children[SELECTCLAUSE].children:
                    if isinstance(child, AggInternal):
                        new_children.append(child)
                    else:
                        gbchild = child
                        gbname = gbchild.value
                        gbattr = GBAttr(gbname)
                        gbattrs.append(gbattr)
                        asgb = As(children=[Node('temp.g['+str(i)+']'), Node(gbname)])
                        orderby = OrderById(children=[Node(gbname)])
                        orderbys.append(orderby)
                        i+=1
                        asgbs.append(asgb)
                node.children[SELECTCLAUSE] = SelectClause(children=new_children)

                select.aggAttr = node.aggAttr

                aggAttr = node.aggAttr

                gbarray = GBArray(children=gbattrs)
                node.children[SELECTCLAUSE].children.append(gbarray)

                child1 = copy.deepcopy(node)
                child2 = copy.deepcopy(node)
                certaincond = BinaryExpression(value='=', children=[Node('trio_enc'), Node('trio_alts')])

                if isinstance(child1.children[WHERECLAUSE], Empty):
                    child1.children[WHERECLAUSE] = WhereClause(children=[certaincond])
                else:
                    child1.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                      children=[Parenthesis(children=[certaincond]),\
                                Parenthesis(children=child1.children[WHERECLAUSE].children)])])

                notcertaincond = BinaryExpression(value='!=', children=[Node('trio_enc'), Node('trio_alts')])

                if isinstance(child2.children[WHERECLAUSE], Empty):
                    child2.children[WHERECLAUSE] = WhereClause(children=[notcertaincond])
                else:
                    child2.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                      children=[Parenthesis(children=[notcertaincond]),\
                                Parenthesis(children=child2.children[WHERECLAUSE].children)])])

                orderid = OrderById(children=[Node('minattr')])
                orderbys.append(orderid)
                child2.children[ORDERBY] = OrderBy(children=orderbys)

                retchild1 = self.visitLavgG1(child1)
                retchild2 = child2

                lavg = Lavgmain(children=[retchild1, retchild2])

                asLavg = As(children=[lavg, Node('temp')])

                select.children.append(SelectClause(children=[]))

                aliasStr = self.getAggAlias('lavg')

                select.children[SELECTCLAUSE].children.append(Node('temp.a as ' + aliasStr))
                for asgb in asgbs:
                    select.children[SELECTCLAUSE].children.append(asgb)

                select.children.append(FromClause(children=[asLavg]))#2
                select.children.append(Empty())#3
                select.children.append(Empty())#4
                select.children.append(Empty())#5
                select.children.append(Empty())#6
                select.children.append(Empty())#7
                select.children.append(Empty())#8

                return select
            child1 = copy.deepcopy(node)
            child2 = copy.deepcopy(node)
            certaincond = BinaryExpression(value='=', children=[Node('trio_enc'), Node('trio_alts')])

            if isinstance(child1.children[WHERECLAUSE], Empty):
                child1.children[WHERECLAUSE] = WhereClause(children=[certaincond])
            else:
                child1.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                  children=[Parenthesis(children=[certaincond]),\
                            Parenthesis(children=child1.children[WHERECLAUSE].children)])])

            notcertaincond = BinaryExpression(value='!=', children=[Node('trio_enc'), Node('trio_alts')])

            if isinstance(child2.children[WHERECLAUSE], Empty):
                child2.children[WHERECLAUSE] = WhereClause(children=[notcertaincond])
            else:
                child2.children[WHERECLAUSE] = WhereClause(children=[BinaryExpression('and',\
                  children=[Parenthesis(children=[notcertaincond]),\
                            Parenthesis(children=child2.children[WHERECLAUSE].children)])])

            orderid = OrderById(children=[Node('minattr')])
            child2.children[ORDERBY] = OrderBy(children=[orderid])

            retchild1 = self.visitLavgQ1(child1)
            retchild2 = child2

            lavg = Lavg(children=[retchild1, retchild2])

            asLavg = self.attachAggAlias(lavg)

            select.children.append(SelectClause(children=[asLavg]))

            select.children.append(Empty())#2
            select.children.append(Empty())#3
            select.children.append(Empty())#4

            retgroupby = self.visit(node.children[GROUPBY])

            select.children.append(retgroupby)

            select.children.append(Empty())#6
            select.children.append(Empty())#7
            select.children.append(Empty())#8

            return select

        if 'emin' == node.aggType:
            select = Select(children=[]);
            select.isConfBase = node.isConfBase
            select.gotAsConf = node.gotAsConf
            select.gotUniform = node.gotUniform
            select.u_columns = node.u_columns
            select.names = node.names
            select.gotAgg = node.gotAgg
            select.aggType = node.aggType

            select.children.append(SelectOptions(children=[Empty()]))

            emmchild = copy.deepcopy(node)
            emmchild.children[GROUPBY] = Empty()

            if node.groupByAgg:
                paren = Emm(children=[emmchild])

                asparen = As(children=[paren,Node('temp')])
                select.children.append(SelectClause(children=[]))

                new_children = []
                gbattrs = []
                asgbs = []
                i=1
                for child in emmchild.children[SELECTCLAUSE].children:
                    if isinstance(child, AggInternal):
                        new_children.append(child)
                    else:
                        gbchild = child
                        gbname = gbchild.value
                        gbattr = GBAttr(gbname)
                        gbattrs.append(gbattr)
                        asgb = As(children=[Node('temp.g['+str(i)+']'), Node(gbname)])
                        i+=1
                        asgbs.append(asgb)

                emmchild.children[SELECTCLAUSE].children = new_children

                gbarray = GBArray(children=gbattrs)

                emmchild.children[SELECTCLAUSE].children.append(gbarray)

                aliasStr = self.getAggAlias('emin')

                select.children[SELECTCLAUSE].children.append(Node('temp.a as ' + aliasStr))

                for asgb in asgbs:
                    select.children[SELECTCLAUSE].children.append(asgb)

                fromclause = FromClause(children=[asparen])

                select.children.append(fromclause)#2
                select.children.append(Empty())#3
                select.children.append(Empty())#4
                select.children.append(Empty())#5
                select.children.append(Empty())#6
                select.children.append(Empty())#7
                select.children.append(Empty())#8

                return select

            paren = Eminmax(children=[emmchild])

            aliasStr = self.getAggAlias('emin')
            asparen = As(children=[paren,Node(aliasStr)])
            select.children.append(SelectClause(children=[asparen]))

            for child in node.children[SELECTCLAUSE].children:
                if not isinstance(child, AggInternal):
                    select.children[SELECTCLAUSE].children.append(child)

            select.children.append(Empty())#2
            select.children.append(Empty())#3
            select.children.append(Empty())#4

            retgroupby = self.visit(node.children[GROUPBY])

            select.children.append(retgroupby)

            select.children.append(Empty())#6
            select.children.append(Empty())#7
            select.children.append(Empty())#8

            return select

        if 'emax' == node.aggType:
            select = Select(children=[]);
            select.isConfBase = node.isConfBase
            select.gotAsConf = node.gotAsConf
            select.gotUniform = node.gotUniform
            select.u_columns = node.u_columns
            select.names = node.names
            select.gotAgg = node.gotAgg
            select.aggType = node.aggType

            select.children.append(SelectOptions(children=[Empty()]))

            if node.groupByAgg:
                emmchild = copy.deepcopy(node)
                emmchild.children[GROUPBY] = Empty()

                paren = Emm(children=[emmchild])

                asparen = As(children=[paren,Node('temp')])
                select.children.append(SelectClause(children=[]))

                new_children = []
                gbattrs = []
                asgbs = []
                i=1
                for child in emmchild.children[SELECTCLAUSE].children:
                    if isinstance(child, AggInternal):
                        new_children.append(child)
                    else:
                        gbchild = child
                        gbname = gbchild.value
                        gbattr = GBAttr(gbname)
                        gbattrs.append(gbattr)
                        asgb = As(children=[Node('temp.g['+str(i)+']'), Node(gbname)])
                        i+=1
                        asgbs.append(asgb)

                emmchild.children[SELECTCLAUSE].children = new_children

                gbarray = GBArray(children=gbattrs)

                emmchild.children[SELECTCLAUSE].children.append(gbarray)

                aliasStr = self.getAggAlias('emax')

                select.children[SELECTCLAUSE].children.append(Node('temp.a as ' + aliasStr))

                for asgb in asgbs:
                    select.children[SELECTCLAUSE].children.append(asgb)

                fromclause = FromClause(children=[asparen])

                select.children.append(fromclause)#2
                select.children.append(Empty())#3
                select.children.append(Empty())#4
                select.children.append(Empty())#5
                select.children.append(Empty())#6
                select.children.append(Empty())#7
                select.children.append(Empty())#8

                return select

            paren = Eminmax(children=[node])
            aliasStr = self.getAggAlias('emax')
            asparen = As(children=[paren,Node(aliasStr)])
            select.children.append(SelectClause(children=[asparen]))

            select.children.append(Empty())#2
            select.children.append(Empty())#3
            select.children.append(Empty())#4

            retgroupby = self.visit(node.children[GROUPBY])

            select.children.append(retgroupby)

            select.children.append(Empty())#6
            select.children.append(Empty())#7
            select.children.append(Empty())#8

            return select

        return node
