__id__ = "$Id$"

from visitor import Visitor
import rewriter
import aggrewriter
import dumper
from trio.node import *
from trio.common import *
import copy

#print debug dumper
def pdebug(node, trio_cursor, debugstr):
    if trio_cursor.debuglevel > 0:
        print debugstr
        dumper.Dumper().visit(node)

#code to prepare refcopy when doing multiple aggs
def prepRefcopy(refcopy, trio_cursor):
    r = rewriter.Rewriter(trio_cursor)
    new_node = r.visit(refcopy)

    #to make all columns of agg tables certain
    new_node.u_columns = []

    return new_node

#code to run rewriter and aggrewriter
def rwAgg(node, trio_cursor, aggalias):
    r = rewriter.Rewriter(trio_cursor)
    r.aggalias = aggalias
    new_node = r.visit(node)

    ar = aggrewriter.AggRewriter(trio_cursor)
    ar.aggalias = aggalias
    new_node = ar.visit(new_node)

    return new_node

def execMultiAggs(node, trio_cursor):
    aggcopy = []
    for i in range(0, len(node.found)):
        nodecopy = copy.deepcopy(node)
        selectaggcopy = copy.deepcopy(node.foundAggs[i])
        nodecopy.children[SELECTCLAUSE] = SelectClause(children=[selectaggcopy])

        # for group by
        try:
            groupByAttrs = node.groupByAttrs
        except AttributeError:
            pass
        else:
            for child in groupByAttrs:
                new_child = copy.deepcopy(child)
                nodecopy.children[SELECTCLAUSE].children.append(new_child)

        aggcopy.append(nodecopy)

        ma = FindAggs(trio_cursor)
        ma.visit(aggcopy[i])

        aggalias = 'trio_agg' + str(i)

        aggcopy[i] = rwAgg(aggcopy[i], trio_cursor, aggalias)

    refcopy = prepRefcopy(copy.deepcopy(node), trio_cursor)

    patcher = PatchMulti(trio_cursor)
    new_node = patcher.patch(refcopy, aggcopy, node)

#    pdebug(new_node, trio_cursor, "after patcher")

    #TODO: line below needed to create table as agg (but a hack)
    #clean this up - look at code in createtableas.py
    node.names = new_node.names

    return new_node

class FindAggs(Visitor):
    def __init__(self, trio_cursor):
        self.trio_cursor = trio_cursor
        self.stack = []

    def visit(self, node):
        self.stack.append(node)
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        new_node = visitfn(node)
        self.stack.pop(-1)
        return new_node

    def visitSelect(self, node):
        self.stack[0].found = []
        self.stack[0].foundAggs = []
        new_children = []
        #for child in node.children:
            #new_children += [self.visit(child)]
        node.children[SELECTCLAUSE] = self.visit(node.children[SELECTCLAUSE])
        self.visit(node.children[GROUPBY])
        return node

    def visitGroupByKey(self, node):
        self.stack[0].groupByAttrs = []
        for child in node.children:
            new_child = copy.deepcopy(child)
            self.stack[0].groupByAttrs += [new_child]
        return node

    def visitDefault(self, node):
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def visitAggregate(self, node):
        aliasStr = node.children[0].__class__.__name__
        self.stack[0].found.append(aliasStr.lower())
        nodecopy = copy.deepcopy(node)
        self.stack[0].foundAggs.append(nodecopy)
        return node

class AggAliasing(Visitor):
    def __init__(self, trio_cursor):
        self.trio_cursor = trio_cursor
        self.stack = []

    def visit(self, node):
        self.stack.append(node)
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        new_node = visitfn(node)
        self.stack.pop(-1)
        return new_node

    def visitSelect(self, node):
        self.stack[0].found = []
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def visitDefault(self, node):
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def visitSelectClause(self, node):
        new_children = []
        for child in node.children:
            if child.__class__.__name__ == 'Aggregate':
                new_children += [self.AttachAlias(child)]
            else:
                new_children += [child]
        node.children = new_children
        return node

    def AttachAlias(self, node):
        aliasStr = node.children[0].__class__.__name__
        aliasNode = Node(aliasStr)
        asAgg = As(children=[node, aliasNode])
        return asAgg

class ReplaceEavg(Visitor):
    def __init__(self, trio_cursor):
        self.trio_cursor = trio_cursor
        self.stack = []

    def visit(self, node):
        self.stack.append(node)
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        new_node = visitfn(node)
        self.stack.pop(-1)
        return new_node

    def visitSelect(self, node):
        self.stack[0].found = []
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def visitDefault(self, node):
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def visitAggregate(self, node):
        if isinstance(node.children[0], Eavg):
            attrName = node.children[0].children[0].value
            esumNode = Esum(children=[Node(attrName)])
            ecountNode = Ecount(children=[Node(attrName)])
            numerator = Aggregate(children=[esumNode])
            denominator = Aggregate(children=[ecountNode])
            quotient = BinaryExpression('/', [numerator, denominator])
            return quotient
        return node

    def visitEavg(self, node):
        return node

class PatchMulti(Visitor):
    def __init__(self, trio_cursor):
        self.trio_cursor = trio_cursor
        self.aggsvisited = 0
        self.stack = []

    def patch(self, refnode, aggnodes, orignode):
        anodes = []

        for i in range(0, len(aggnodes)):
            pnode = Parenthesis(children=[aggnodes[i]])
            anode = As(children=[pnode, Node('s'+str(i))])
            anodes.append(anode)

        origsc = copy.deepcopy(orignode.children[SELECTCLAUSE])
        refnode.children[SELECTCLAUSE]=origsc

        self.visit(refnode.children[SELECTCLAUSE])

        refnode.children[FROMCLAUSE]=FromClause()
        refnode.children[FROMCLAUSE].children=anodes

        # join on group by attr
        try:
            groupByAttrs = orignode.groupByAttrs
        except AttributeError:
            refnode.children[WHERECLAUSE] = Empty()
        else:
            refnode.children[WHERECLAUSE] = Empty()
            if len(aggnodes)>1:
                wherecond=Empty()
                for child in groupByAttrs:
                    gbAttrName = child.value
                    for i in range(0, len(aggnodes)-1):
#                    stri0 = 's'+str(i)+'.'+gbAttrName
#                    stri1 = 's'+str(i+1)+'.'+gbAttrName
                        # cast to text so output from stored procedures can join as well
                        stri0 = 'cast(s'+str(i)+'.'+gbAttrName + ' as text)'
                        stri1 = 'cast(s'+str(i+1)+'.'+gbAttrName + ' as text)'

                        newcond = BinaryExpression(value='=', children=[Node(stri0), Node(stri1)])
                        if isinstance(wherecond, Empty):
                            wherecond = newcond
                        else:
                            wherecond = BinaryExpression('and',\
                              children=[Parenthesis(children=[newcond]),\
                                        Parenthesis(children=[wherecond])])
                refnode.children[WHERECLAUSE] = WhereClause(children=[wherecond])

        refnode.children[GROUPBY] = Empty()
        refnode.children[ORDERBY] = Empty()

        self.appendAggTrioColumns(refnode.children[SELECTCLAUSE])

        return refnode

    def visit(self, node):
        self.stack.append(node)
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        new_node = visitfn(node)
        self.stack.pop(-1)
        return new_node

    def visitSelect(self, node):
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    # disambiguate group by attr
    def visitSelectClause(self, node):
        new_children = []
        for child in node.children:
            if child.__class__.__name__ == 'Node':
                child.value = 's0.' + child.value
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def visitDefault(self, node):
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node

    def visitAggregate(self, node):
        aliasStr = 'trio_agg' + str(self.aggsvisited)
        node.children=[Node(aliasStr)]
        self.aggsvisited += 1
        return node

    def appendAggTrioColumns(self, node):
        node.children = node.children + \
        [As(children=[Node("nextval('aggs_trio_xid')"), Node("trio_xid")])] + \
        [As(children=[Node("nextval('aggs_trio_aid')"), Node("trio_aid")])] + \
        [As(children=[Node("1"), Node("trio_alts")])] + \
        [As(children=[Node("1"), Node("trio_enc")])] + \
        [As(children=[Node("1"), Node('trio_conf')])]
        return node
