__id__ = "$Id$"

import rewriter
from trio.node import *
from visitor import Visitor

class Restructurer(Visitor):
    # For the restructurer a Horizontal Subquery (HS) is 'just' a regular
    # item. The context is not changed when a HS node is entered.
    def __init__(self):
        self.context = []
        self.tupleProducing = False

    def visit(self, node):
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        return visitfn(node)

    def visitDefault(self, node):
        #print node
        #print len(node.children)
        for n in node.children:
            self.visit(n)
        return node

    def visitSelectClause(self, node):
        tupleProducing = False
        for n in node.children:
            if isinstance(n, HorizontalSelect):
                if not isinstance(n.children[3], Empty):
                    raise Exception("GROUP-BY not allowed for horizontal subqueries in SELECT!")
                if (len(n.children[0].children) == 1 and \
                    not isinstance(n.children[0].children[0], Aggregate)) \
                   or len(n.children[0].children) > 1 \
                   or isinstance(n.children[0], Star):
                    tupleProducing = True
                rewriter.checkAliases(n, True)
            self.visit(n)
        self.tupleProducing = tupleProducing
        return node

    def visitWhereClause(self, node):
        if self.tupleProducing:
            raise Exception("WHERE not allowed in outer query with tuple-producing subquery!")
        for n in node.children:
            self.visit(n)
        return node

    # "Lift" the horizontal select and where clauses and reconstruct the outer query
    def visitSelect(self, node):
        for n in node.children:
            self.visit(n)
        if self.tupleProducing:
            if len(node.children[1].children) > 1:
                raise Exception("Tuple-producing subquery must be the only expression in SELECT!")
            rewriter.checkAliases(node)
            if not isinstance(node.children[1].children[0].children[1], Empty):
                self.checkSubset(node.children[1].children[0].aliasdict, node.aliasdict)
                self.checkSubset(node.aliasdict, node.children[1].children[0].aliasdict, False)
            self.switchFromAliases(node.children[1].children[0].children[0], node.children[1].children[0], node)
            self.switchFromAliases(node.children[1].children[0].children[2], node.children[1].children[0], node)
            node.aliasdict = node.children[1].children[0].aliasdict
            new_children = [node.children[0]]
            new_children.append(node.children[1].children[0].children[0])
            new_children.append(node.children[2])
            new_children.append(node.children[1].children[0].children[2])
            new_children.append(node.children[4])
            new_children.append(node.children[5])
            new_children.append(node.children[6])
            new_children.append(node.children[7])
            node.children = new_children
        return node

    def checkSubset(self, fromAliases, toAliases, outer=True):
        for from_alias, from_name in fromAliases.iteritems():
            matched = False
            for to_alias, to_name in toAliases.iteritems():
                if to_name == from_name:
                    matched = True
                    break
            if not matched:
                if outer:
                    raise Exception("Unresolvable table reference '%s' in outer query" % from_name)
                else:
                    raise Exception("Unresolvable table reference '%s' in horizontal query" % from_name)

    # Apply outer table aliases to lifted horizontal query clauses.
    # For self-joins, the i_th table alias in the inner query has to match
    # the i_th table alias in the outer query with the same table name
    # e.g.: select [select a.type, b.type from crime a, crime b where a.type = b.type] from crime c, crime d;
    def switchFromAliases(self, node, fromNode, toNode):
        for n in node.children:
            if isinstance(n, Dot):
                if fromNode.aliasdict.has_key(n.children[0].value + "_inner"):
                    from_alias = n.children[0].value + "_inner"
                    from_tablename = fromNode.aliasdict[from_alias]
                elif toNode.aliasdict.has_key(n.children[0].value):
                    from_alias = n.children[0].value
                    from_tablename = toNode.aliasdict[from_alias]
                else:
                    raise Exception("Unresolvable alias or table reference '%s' in horizontal query" % n.children[0].value)
                i = 0
                j1 = 0
                for name in fromNode.names:
                    if from_tablename == name:
                        if from_alias != fromNode.aliases[i]:
                            j1 += 1
                        else:
                            break
                    i += 1
                i = 0
                j2 = 0
                for to_tablename in toNode.names:
                    if to_tablename == from_tablename:
                        if j1 == j2:
                            n.children[0].value = toNode.aliases[i]
                            break
                        else:
                            j2 += 1
                    i += 1
            else:
                self.switchFromAliases(n, fromNode, toNode)
