__id__ = "$Id$"

from trio.node import *
from visitor import Visitor

# This class prints TriQL parse trees back to TriQL query strings
# (needed for manipulating subqueries etc. at the TriQL level)
#
# Currently, this class is only used by Rewriter.visitSetOperator().
# Thus, it only needs to print TriQL select query.

class TriQLPrinter(Visitor):
    def visit(self, node):
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        return visitfn(node)

    def visitDefault(self, node):
        return str(node.value)

    def visitCreateIndex(self, node): pass
    def visitCreateTable(self, node): pass
    def visitCreateTempTable(self, node): pass
    def visitCreateTableAs(self, node): pass
    def visitCreateTempTableAs(self, node): pass
    def visitDropIndex(self, node): pass
    def visitDropTable(self, node): pass
    def visitInsert(self, node): pass
    def visitSetOperator(self, node): pass

    def visitSelect(self, node):
        sql = "Select " + self.visit(node.children[0]) + " " +\
              self.visit(node.children[1]) + \
              self.visit(node.children[2]) + \
              self.visit(node.children[3]) + \
              self.visit(node.children[4]) + \
              self.visit(node.children[5]) + \
              self.visit(node.children[6]) + \
              self.visit(node.children[7])
        return sql

    def visitSelectOptions(self, node):
        return " ".join([self.visit(x) for x in node.children])

    def visitSelectClause(self, node):
        return ", ".join([self.visit(x) for x in node.children])

    def visitFromClause(self, node):
        return " From " + ", ".join([self.visit(x) for x in node.children])

    def visitWhereClause(self, node):
        return " Where " + self.visit(node.children[0])


    def visitGroupByKey(self, node):
        return " Group By " + ", ".join([self.visit(x) for x in node.children])

    def visitOrderBy(self, node):
        return " Order By " + ", ".join([self.visit(x) for x in node.children])

    def visitComputeConfidences(self, node):
        return " Compute Confidences"

    def visitLiteral(self, node):
        if isinstance(node.value, str):
            return "'%s'" % node.value
        elif isinstance(node.value, unicode):
            return "'%s'" % node.value.encode('utf-8')
        elif node.value is None:
            return "null"
        else:
            return str(node.value)

    def visitBinaryExpression(self, node):
        return self.visit(node.children[0]) + \
               " " + str(node.value) + " " + \
               self.visit(node.children[1])

    def visitUnaryExpression(self, node):
        return str(node.value) + self.visit(node.children[0])

    def visitStar(self, node):
        return "*"

    def visitAs(self, node):
        return self.visit(node.children[0]) + " As " + \
               self.visit(node.children[1])

    def visitDot(self, node):
        return self.visit(node.children[0]) + "." + \
               self.visit(node.children[1])

    def visitEmpty(self, node):
        return ""

    def visitConfAgg(self, node):
        return ""

    def visitConfUniform(self, node):
        return "uniform as conf"

    def visitConf(self, node):
        return "conf(" + self.visit(node.children[0]) + ")"

    def visitLineage(self, node):
        return self.visit(node.children[0]) + " ==> " + self.visit(node.children[1])

    def visitParenthesis(self, node):
        return "(" + self.visit(node.children[0]) + ")"

    def visitGroupAlts(self, node):
        return "groupalts(" + self.visit(node.children[0]) + ")"

    def visitHorizontalSelect(self, node):
        return " [ " + self.visit(node.children[0]) + \
            self.visit(node.children[1]) + \
            self.visit(node.children[2]) + \
            self.visit(node.children[3]) + \
            self.visit(node.children[4])+ " ] "

    def visitGroupbyClause(self, node):
        return " Group By " + ", ".join([self.visit(x) for x in node.children])

    def visitHavingClause(self, node):
        return " Having " + self.visit(node.children[0])

    def visitColumn(self, node):
        return self.visit(node.children[0]) + " " + self.visit(node.children[1])

    def visitMaybe(self, node):
        return "maybe(" + str(node.value) + ")"

    def visitColumnList(self, node):
        return ", ".join([self.visit(x) for x in node.children])

    def visitAlternative(self, node):
        return ", ".join([self.visit(x) for x in node.children])

    def visitDataType(self, node):
        if len(node.children) == 0 or isinstance(node.children[0], Empty):
            return node.value
        else:
            return node.value + "(" + self.visit(node.children[0]) + ")"

    def visitUniformTable(self, node):
        return "uniform " + node.value

    def visitFunction(self, node):
        return "%s(%s)" % (node.value,  ", ".join([self.visit(n) for n in node.children]))

    def visitAggregate(self, node):
        return self.visit(node.children[0]) # visit the actual aggregate

    def visitMin(self, node):
        return "min(" + self.visit(node.children[0]) + ")"

    def visitMax(self, node):
        return "max(" + self.visit(node.children[0]) + ")"

    def visitAvg(self, node):
        return "avg(" + self.visit(node.children[0]) + ")"

    def visitSum(self, node):
        return "sum(" + self.visit(node.children[0]) + ")"

    def visitCount(self, node):
        return "count(" + self.visit(node.children[0]) + ")"

    def visitLmin(self, node):
        return "lmin(" + self.visit(node.children[0]) + ")"

    def visitLmax(self, node):
        return "lmax(" + self.visit(node.children[0]) + ")"

    def visitLavg(self, node):
        return "lavg(" + self.visit(node.children[0]) + ")"

    def visitLsum(self, node):
        return "lsum(" + self.visit(node.children[0]) + ")"

    def visitLcountStar(self, node):
        return "lcount(*)"

    def visitHmin(self, node):
        return "hmin(" + self.visit(node.children[0]) + ")"

    def visitHmax(self, node):
        return "hmax(" + self.visit(node.children[0]) + ")"

    def visitHavg(self, node):
        return "havg(" + self.visit(node.children[0]) + ")"

    def visitHsum(self, node):
        return "hsum(" + self.visit(node.children[0]) + ")"

    def visitHcountStar(self, node):
        return "hcount(*)"

    def visitEmin(self, node):
        return "emin(" + self.visit(node.children[0]) + ")"

    def visitEmax(self, node):
        return "emax(" + self.visit(node.children[0]) + ")"

    def visitEavg(self, node):
        return "eavg(" + self.visit(node.children[0]) + ")"

    def visitEsum(self, node):
        return "esum(" + self.visit(node.children[0]) + ")"

    def visitEcount(self, node):
        return "ecount(" + self.visit(node.children[0]) + ")"

    def visitEcountStar(self, node):
        return "ecount(*)"
